"""Configuration handler.

This script provides the base class to save the metadata for a given ReEDS scenario.
It can either read the information directly or throught a cases file.
"""

from dataclasses import field
from enum import StrEnum
from typing import Any, TypeVar

from pydantic import BaseModel

BaseModelConfigType = TypeVar("BaseModelConfigType", bound=BaseModel)


class BaseModelConfig(BaseModel):
    """Configuration of the Model.

    Attributes
    ----------
    fmap
        Dictionary with file to be parse
    defaults
        Default configuration for the model

    Raises
    ------
    AttributeError
    """

    defaults: dict[str, Any] = field(default_factory=dict)
    fmap: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def get_field_mapping(cls) -> dict[type[BaseModel], dict[str, str]]:
        """Return a dict of {target_class: {target_field: source_field}}."""
        raise NotImplementedError

    def transform_field(self, field_name: str, target_class: type[BaseModel]) -> Any:
        """Define custom field transformations."""
        return getattr(self, field_name)

    def to_class(
        self, target_class: type[BaseModelConfigType], base_instance: BaseModel | None = None
    ) -> BaseModelConfigType:
        """Transform this model instance to another Pydantic model."""
        field_mappings = self.get_field_mapping().get(target_class, {})

        transformed_data = base_instance.model_dump() if base_instance is not None else {}

        for target_field, source_field in field_mappings.items():
            transformed_data[target_field] = self.transform_field(source_field, target_class)

        return target_class(**transformed_data)


class ReEDSConfig(BaseModelConfig):
    """ReEDs specific configuration."""

    solve_year: list[int] | int | None = None
    weather_year: int | None = None

    @classmethod
    def get_field_mapping(cls) -> dict[type[BaseModel], dict[str, str]]:
        """Return a dict of {target_class: {target_field: source_field}}."""
        return {
            PlexosConfig: {
                "model_year": "solve_year",
                "horizon_year": "weather_year",
            },
            SiennaConfig: {"model_year": "solve_year"},
        }


class PlexosConfig(BaseModelConfig):
    """Plexos specific configuration."""

    master_file: str | None = None
    model_name: str | None = None
    model_year: int | None = None
    horizon_year: int | None = None


class SiennaConfig(BaseModelConfig):
    """Sienna specific configuration."""

    model_year: int | None = None


class InfrasysConfig(BaseModelConfig):
    """Infrasys specific configuration."""

    model_year: int | None = None

    @classmethod
    def get_field_mapping(cls) -> dict[type[BaseModel], dict[str, str]]:
        """Return a dict of {target_class: {target_field: source_field}}."""
        return {
            PlexosConfig: {
                "model_year": "model_year",
            },
            SiennaConfig: {"model_year": "model_year"},
        }


class Models(StrEnum):
    """Enum of valid models supported."""

    INFRASYS = "INFRASYS"
    REEDS = "REEDS-US"
    PLEXOS = "PLEXOS"
    SIENNA = "SIENNA"
    PRAS = "PRAS"


class ParserModels(StrEnum):
    """Enum of valid parser models supported."""

    INFRASYS = "INFRASYS"
    REEDS = "REEDS-US"
    PLEXOS = "PLEXOS"
    SIENNA = "SIENNA"


class ExporterModels(StrEnum):
    """Enum of valid exporter models supported."""

    PLEXOS = "PLEXOS"
    SIENNA = "SIENNA"
    INFRASYS = "INFRASYS"


MODEL_CONFIGS = {
    Models.REEDS: ReEDSConfig,
    Models.PLEXOS: PlexosConfig,
    Models.SIENNA: SiennaConfig,
    Models.INFRASYS: InfrasysConfig,
}



"""R2X CLI.

This scripts contains most of the logic on how we call different parts of R2X using the CLI.
"""

import sys
import traceback

from .cli_functions import base_cli
from .logger import setup_logging
from .runner import init, run
from .utils import read_user_dict


def cli() -> None:
    """CLI main entry point for R2X."""
    parser = base_cli()
    args, remaining_args = parser.parse_known_args()

    if "--help" in remaining_args or len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(0)

    args = parser.parse_args()

    cli_args = {k: v for k, v in vars(args).items() if v is not None}
    user_dict = cli_args.pop("user_config", None)
    setup_logging(verbosity=cli_args["verbose"])

    if user_dict is not None:
        user_dict = read_user_dict(user_dict)

    if cli_args.get("pdb"):
        import pdb

        try:
            pdb.run("cli_commands(cli_args, user_dict)", globals=globals(), locals=locals())
        except Exception:
            traceback.print_exc()
            pdb.post_mortem()
    else:
        cli_commands(cli_args, user_dict)
    return None


def cli_commands(cli_args: dict, user_dict: dict | None = None) -> None:
    """Run the different translation for a configuration.

    If the config object is Configuration class, we run all the different
    Scenarios in parallel using multiprocessing.

    Parameters
    ----------
    config
        Configuration

    Other Parameters
    ----------------
    kwargs
        arguments passed for convenience.
    """
    if cli_args["command"] == "run":
        run(cli_args, user_dict=user_dict)
    elif cli_args["command"] == "init":
        init(cli_args)
    else:
        raise NotImplementedError
    return



"""R2X API for data model."""

import csv
from collections.abc import Callable
from os import PathLike
from pathlib import Path
from collections.abc import Iterable
from loguru import logger

import inspect
from infrasys.component import Component
from infrasys.system import System as ISSystem

from .__version__ import __data_model_version__


class System(ISSystem):
    """API to interact with the SystemModel."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_format_version = __data_model_version__

    def __str__(self) -> str:
        return f"System(name={self.name}, DataModel Version={self.version})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def version(self):
        """The version property."""
        return __data_model_version__

    def to_json(self, filename: Path | str, overwrite=False, indent=None, data=None) -> None:  # noqa: D102
        return super().to_json(filename, overwrite=overwrite, indent=indent, data=data)

    @classmethod
    def from_json(cls, filename: Path | str, upgrade_handler: Callable | None = None, **kwargs) -> "System":  # noqa: D102
        return super().from_json(filename=filename, upgrade_handler=upgrade_handler, **kwargs)  # type: ignore

    def export_component_to_csv(
        self,
        component: type[Component],
        fields: list | None = None,
        filter_func: Callable | None = None,
        fpath: PathLike | None = None,
        key_mapping: dict | None = None,
        unnest_key: str = "name",
        **dict_writer_kwargs,
    ):
        """Export components into a csv.

        component:
            Component type to get from the system
        """
        # Get desired components to offload to csv
        components = map(
            lambda component: component.model_dump(
                exclude={},
                exclude_none=True,
                mode="json",
                context={"magnitude_only": True},
                # serialize_as_any=True,
            ),
            self.get_components(component, filter_func=filter_func),
        )
        if fpath is None:
            fpath = Path(f"{component.__name__}.csv")

        self._export_dict_to_csv(
            components,
            fpath=fpath,
            fields=fields,
            key_mapping=key_mapping,
            unnest_key=unnest_key,
            **dict_writer_kwargs,
        )

    def _export_dict_to_csv(
        self,
        data: Iterable[dict],
        fpath: PathLike,
        fields: list | None = None,
        # key_mapping: dict | None = None,
        # unnest_key: str = "name",
        **dict_writer_kwargs,
    ):
        dict_writer_kwargs = {
            key: value
            for key, value in dict_writer_kwargs.items()
            if key in inspect.getfullargspec(csv.DictWriter).args
        }

        with open(str(fpath), "w", newline="") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fields, extrasaction="ignore", **dict_writer_kwargs)  # type: ignore
            writer.writeheader()
            for row in data:
                writer.writerow(row)
        return


if __name__ == "__main__":
    from .logger import setup_logging
    from rich.console import Console

    setup_logging(level="TRACE")
    logger.enable("infra_sys")

    console = Console()

    # from tests.models.systems import ieee_5bus
    #
    # system = ieee_5bus()
    # logger.info("From ARTEX")



"""Exporter base class."""

from abc import ABC, abstractmethod
import string
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pandas as pd
import numpy as np
import infrasys
from loguru import logger
from pint import Quantity

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.exporter.utils import modify_components
from r2x.parser.handler import file_handler

OUTPUT_FNAME = "{self.weather_year}"


class BaseExporter(ABC):
    """Class that defines the shared methods of parsers.

    Note
    ----
    This class is meant to be use for developing new parsers. Do not use it directly.

    Attributes
    ----------
    config:
        Scenario configuration
    system:
        System class
    output_foder:
        output folder path

    Methods
    -------
    run()
        Sequentially run methods from the exporter
    get_data(key='load')
        Return the parsed data for load.
    read_file(fpath="load.csv")
        Read load data.
    export_data_files(time_series_folder="Data")
        Save time series data files.
    """

    def __init__(
        self, config: Scenario, system: System, output_folder: Path, ts_directory: str = "Data"
    ) -> None:
        self.config = config
        self.system = system
        self.output_folder = output_folder
        self.input_model = config.input_model
        self.ts_directory = Path(ts_directory)
        self.time_series_objects: dict[str, list[Any]] = defaultdict(list)
        self.time_series_name_by_type: dict[str, list[Any]] = defaultdict(list)
        self._handle_data_folder(config.output_folder, self.ts_directory)

    def _handle_data_folder(self, output_folder: str | Path, folder_name: str | Path) -> None:
        fpath = Path(output_folder) / folder_name
        if not fpath.exists():
            fpath.mkdir()
        return

    @abstractmethod
    def run(self, *args, **kwargs) -> "BaseExporter":
        """Run the exporter methods in sequence.

        This methods needs to be implemented by any exporter and it's content
        change based on the order the data is being exported.
        """

    def read_file(self, fpath: Path | str, filter_func: list[Callable] | None = None, **kwargs):
        """Read input model data from the file system.

        Currently supported formats:
            - .csv
            - .h5
        More to come!

        Parameters
        ----------
        fpath: Path, str
            Absolute location of the file in the system
        filter_func: List
            Filter functions to apply

        """
        data = file_handler(fpath, **kwargs)
        if data is None:
            return

        if isinstance(filter_func, list):
            for func in filter_func:
                data = func(data, **kwargs)
        return data

    def export_data_files(self, year: int, time_series_folder: str = "Data") -> None:
        """Export all time series objects attached to components.

        This method assumes that `self.config.weather_year and `self.output_folder` exist.

        Parameters
        ----------
        time_series_folder: str
            Folder name to save time series data
        """
        assert year is not None
        config_dict = self.config.__dict__
        config_dict["year"] = year
        for component in self.system.iter_all_components():
            if self.system.has_time_series(component):
                for ts_metadata in self.system.time_series.list_time_series_metadata(component):
                    ts_component_name = f"{component.__class__.__name__}_{ts_metadata.variable_name}"
                    try:
                        self.time_series_objects[ts_component_name].append(
                            self.system.get_time_series(component, variable_name=ts_metadata.variable_name)
                        )
                    except:  # noqa: E722
                        continue
                    self.time_series_name_by_type[ts_component_name].append(component.name)

        component_lengths = {
            component_type: {ts.length}
            for component_type, time_series in self.time_series_objects.items()
            for ts in time_series
        }

        inconsistent_lengths = [
            (component_type, length_set)
            for component_type, length_set in component_lengths.items()
            if len(length_set) != 1
        ]
        if inconsistent_lengths:
            raise ValueError(f"Multiple lengths found for components time series: {inconsistent_lengths}")

        datetime_arrays = {
            component_type: (
                np.datetime_as_string(
                    pd.date_range(
                        start=f"1/1/{year}",
                        periods=ts.length,
                        freq=f"{int(ts.resolution.total_seconds() / 60)}min",  # Convert resolution to minutes
                    ),
                    unit="m",
                ),
                time_series,
            )
            for component_type, time_series in self.time_series_objects.items()
            for ts in time_series
        }
        csv_fpath = self.output_folder / time_series_folder

        # Use string substitution to dynamically change the output csv fnames
        csv_fname = config_dict.get("time_series_fname", "${component_type}_${name}_${year}.csv")
        logger.trace("Using {} as time_series name", csv_fname)
        string_template = string.Template(csv_fname)

        for component_type, (datetime_array, time_series) in datetime_arrays.items():
            time_series_arrays = list(
                map(lambda x: x.data.magnitude if isinstance(x.data, Quantity) else x.data, time_series)
            )
            config_dict["component_type"] = component_type
            csv_fname = string_template.safe_substitute(config_dict)
            csv_table = np.column_stack([datetime_array, *time_series_arrays])
            header = '"DateTime",' + ",".join(
                [f'"{name}"' for name in self.time_series_name_by_type[component_type]]
            )

            np.savetxt(
                csv_fpath / csv_fname,
                csv_table,
                header=header,
                delimiter=",",
                comments="",
                fmt="%s",
            )

        return


def get_export_records(component_list: list[dict[str, Any]], *update_funcs: Callable) -> list[dict[str, Any]]:
    """Apply update functions to a list of components and return the modified list.

    Parameters
    ----------
    component_list : list[dict[str, Any]]
        A list of dictionaries representing components to be updated.
    *update_funcs : Callable
        Variable number of update functions to be applied to each component.

    Returns
    -------
    list[dict[str, Any]]
        A list of updated component dictionaries.

    Examples
    --------
    >>> def update_name(component):
    ...     component["name"] = component["name"].upper()
    ...     return component
    >>> def add_prefix(component):
    ...     component["id"] = f"PREFIX_{component['id']}"
    ...     return component
    >>> components = [{"id": "001", "name": "Component A"}, {"id": "002", "name": "Component B"}]
    >>> updated_components = get_export_records(components, update_name, add_prefix)
    >>> updated_components
    [{'id': 'PREFIX_001', 'name': 'COMPONENT A'}, {'id': 'PREFIX_002', 'name': 'COMPONENT B'}]
    """
    update_functions = modify_components(*update_funcs)
    return [update_functions(component) for component in component_list]


def get_export_properties(component, *update_funcs: Callable) -> dict[str, Any]:
    """Apply update functions to a single component and return the modified component.

    Parameters
    ----------
    component : dict[str, Any]
        A dictionary representing a component to be updated.
    *update_funcs : Callable
        Variable number of update functions to be applied to the component.

    Returns
    -------
    dict[str, Any]
        The updated component dictionary.

    Examples
    --------
    >>> def update_status(component):
    ...     component["status"] = "active"
    ...     return component
    >>> def add_timestamp(component):
    ...     from datetime import datetime
    ...
    ...     component["last_updated"] = datetime.now().isoformat()
    ...     return component
    >>> component = {"id": "003", "name": "Component C"}
    >>> updated_component = get_export_properties(component, update_status, add_timestamp)
    >>> updated_component
    {'id': '003', 'name': 'Component C', 'status': 'active', 'last_updated': '2024-09-27T10:30'}
    """
    update_functions = modify_components(*update_funcs)
    return update_functions(component)


def get_exporter(
    config: Scenario,
    system: infrasys.system.System,
    exporter_class,
    filter_funcs: list[Callable] | None = None,
    **kwargs,
) -> BaseExporter:
    """Return exporter instance.


    Parameters
    ----------
    config
        Scenario configuration class
    system
        Infrasys system
    filter_func:
        Functions that will applied to read_data process

    Other Parameters
    ----------------
    kwargs
        year
            For filtering by solve year
        year_column
            To change the column to apply the filter
        column_mapping
            For renaming columns

    See Also
    --------
    BaseExporter
    pl_filter_year
    pl_lower_case
    pl_rename
    """
    logger.debug("Getting {} instance.", exporter_class.__name__)

    exporter: BaseExporter = exporter_class(
        config=config, system=system, output_folder=config.output_folder, **kwargs
    )

    # Functions relative to the parser.
    # NOTE: At some point we are going to migrate this out, but this sound like a good standard set
    # if filter_funcs is None:
    #     logger.trace("Using default filter functions")
    #     filter_funcs = [pl_lowercase, pl_rename, pl_filter_year]

    exporter.run(
        config=config,
        filter_func=filter_funcs,
        **{**config.input_config.__dict__, **kwargs},
    )

    return exporter



import pytest
from r2x.config_scenario import Scenario
from r2x.exporter.plexos import PlexosExporter
from r2x.parser.handler import get_parser_data
from r2x.parser.reeds import ReEDSParser


@pytest.fixture
def scenario_instance(reeds_data_folder, default_scenario, tmp_folder):
    return Scenario.from_kwargs(
        name=default_scenario,
        input_model="reeds-US",
        run_folder=reeds_data_folder,
        output_model="plexos",
        output_folder=tmp_folder,
        solve_year=2050,
        weather_year=2012,
    )


@pytest.fixture
def reeds_parser_instance(scenario_instance):
    return get_parser_data(scenario_instance, parser_class=ReEDSParser)


@pytest.fixture
def reeds_system(reeds_parser_instance):
    return reeds_parser_instance.build_system()


@pytest.fixture
def plexos_exporter(scenario_instance, reeds_system, tmp_folder):
    return PlexosExporter(config=scenario_instance, system=reeds_system, output_folder=tmp_folder)


@pytest.mark.plexos
def test_plexos_exporter_instance(plexos_exporter):
    assert isinstance(plexos_exporter, PlexosExporter)


@pytest.mark.plexos
def test_plexos_exporter_run(plexos_exporter, default_scenario, tmp_folder):
    exporter = plexos_exporter.run()

    output_files = [
        f"{default_scenario}.xml",
    ]

    for file in output_files:
        assert (tmp_folder / file).exists()

    # Check that time series was created correctly
    ts_directory = tmp_folder / exporter.ts_directory
    assert any(ts_directory.iterdir())


@pytest.mark.plexos
def test_plexos_operational_cost(reeds_system, plexos_exporter): ...



"""Functions related to parsers."""

import importlib
from argparse import ArgumentParser
from collections import defaultdict
from datetime import datetime, timedelta
from itertools import repeat
from operator import attrgetter

import numpy as np
import polars as pl
import pyarrow as pa
from infrasys.cost_curves import CostCurve, FuelCurve, UnitSystem
from infrasys.function_data import LinearFunctionData
from infrasys.time_series_models import SingleTimeSeries
from infrasys.value_curves import AverageRateCurve, LinearCurve
from loguru import logger
from pint import Quantity

from r2x.api import System
from r2x.config_models import ReEDSConfig
from r2x.enums import ACBusTypes, EmissionType, PrimeMoversType, ReserveDirection, ReserveType, ThermalFuels
from r2x.exceptions import ParserError
from r2x.models import (
    ACBus,
    Area,
    Bus,
    Emission,
    Generator,
    GenericBattery,
    HybridSystem,
    HydroGen,
    LoadZone,
    MonitoredLine,
    PowerLoad,
    RenewableDispatch,
    RenewableNonDispatch,
    Reserve,
    ReserveMap,
    TransmissionInterface,
    TransmissionInterfaceMap,
)
from r2x.models.core import MinMax
from r2x.models.costs import HydroGenerationCost, ThermalGenerationCost
from r2x.models.generators import HydroDispatch, HydroEnergyReservoir, RenewableGen, ThermalGen
from r2x.parser.handler import BaseParser, create_model_instance
from r2x.units import ActivePower, EmissionRate, Energy, Percentage, Time, ureg
from r2x.utils import get_enum_from_string, match_category, read_csv

from .polars_helpers import pl_left_multi_join

R2X_MODELS = importlib.import_module("r2x.models")
UNITS = importlib.import_module("r2x.units")
BASE_WEATHER_YEAR = 2007


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--weather-year",
        type=int,
        dest="weather_year",
        help="ReEDS weather year to translate",
    )


class ReEDSParser(BaseParser):
    """ReEDS parser class."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert self.config.input_config
        assert isinstance(self.config.input_config, ReEDSConfig)
        self.reeds_config = self.config.input_config
        if not self.reeds_config.weather_year:
            raise AttributeError("Missing weather year from the configuration class.")
        if not self.reeds_config.solve_year:
            raise AttributeError("Missing solve year from the configuration class.")
        self.device_map = self.reeds_config.defaults["reeds_device_map"]
        self.tech_to_fuel_pm = self.reeds_config.defaults["tech_to_fuel_pm"]
        self.excluded_categories = self.reeds_config.defaults["excluded_categories"]
        self.weather_year: int = self.reeds_config.weather_year
        self.skip_validation: bool = getattr(self.reeds_config, "skip_validation", False)

        # Add hourly_time_index
        self.hourly_time_index = np.arange(
            f"{self.weather_year}",
            f"{self.weather_year + 1}",
            dtype="datetime64[h]",
        )[:-24]  # Removing 1 day to match ReEDS convention and converting into a vector
        self.daily_time_index = np.arange(
            f"{self.weather_year}",
            f"{self.weather_year + 1}",
            dtype="datetime64[D]",
        )[:-1]  # Removing 1 day to match ReEDS convention and converting into a vector

    def build_system(self) -> System:
        """Create IS system for the ReEDS model."""
        self.system = System(name=self.config.name, auto_add_composed_components=True)

        # Construct transmission network and buses
        self._construct_buses()
        self._construct_reserves()
        self._construct_branches()
        self._construct_tx_interfaces()
        self._construct_generators()

        # Construct additional objects
        self._construct_emissions()

        # Time series construction
        self._construct_load()
        self._construct_hydro_budgets()
        self._construct_hydro_rating_profiles()
        self._construct_cf_time_series()
        self._construct_reserve_provision()
        self._construct_hybrid_systems()

        return self.system

    # NOTE: Rename to create topology
    def _construct_buses(self):
        logger.info("Creating bus objects.")
        bus_data = self.get_data("hierarchy")

        zones = bus_data["transmission_region"].unique()
        for zone in zones:
            self.system.add_component(self._create_model_instance(LoadZone, name=zone))

        for area in bus_data["state"].unique():
            self.system.add_component(self._create_model_instance(Area, name=area))

        for idx, bus in enumerate(bus_data.iter_rows(named=True)):
            self.system.add_component(
                self._create_model_instance(
                    ACBus,
                    number=idx + 1,
                    name=bus["region"],
                    area=self.system.get_component(Area, name=bus["state"]),
                    load_zone=self.system.get_component(LoadZone, name=bus["transmission_region"]),
                    bus_type=ACBusTypes.PV,
                )
            )

    def _construct_reserves(self):
        logger.info("Creating reserves objects.")
        bus_data = self.get_data("hierarchy")

        reserves = bus_data["transmission_region"].unique()
        for reserve in reserves:
            for name in self.reeds_config.defaults["default_reserve_types"]:
                reserve_duration = self.reeds_config.defaults["reserve_duration"].get(name)
                time_frame = self.reeds_config.defaults["reserve_time_frame"].get(name)
                load_risk = self.reeds_config.defaults["reserve_load_risk"].get(name)
                vors = self.reeds_config.defaults["reserve_vors"].get(name)
                reserve_area = self.system.get_component(LoadZone, name=reserve)
                self.system.add_component(
                    self._create_model_instance(
                        Reserve,
                        name=f"{reserve}_{name}",
                        region=reserve_area,
                        reserve_type=ReserveType[name],
                        vors=vors,
                        duration=reserve_duration,
                        load_risk=load_risk,
                        time_frame=time_frame,
                        direction=ReserveDirection.UP,
                    )
                )
        # Add reserve map
        self.system.add_component(self._create_model_instance(ReserveMap, name="reserve_map"))

    def _construct_branches(self):
        logger.info("Creating branch objects.")
        branch_data = self.get_data("tx_cap")
        tx_loss = self.get_data("tx_losses")

        branch_data = pl_left_multi_join(branch_data, tx_loss)
        ext = {"Wheeling Charge": 0.001, "Wheeling Charge Back": 0.001}
        reverse_lines = set()
        for idx, branch in enumerate(branch_data.iter_rows(named=True)):
            from_bus = self.system.get_component(ACBus, branch["from_bus"])
            to_bus = self.system.get_component(ACBus, branch["to_bus"])
            branch_name = f"{idx + 1:>04}-{branch['from_bus']}-{branch['to_bus']}"
            reverse_key = (branch["kind"], branch["from_bus"], branch["to_bus"])
            if reverse_key in reverse_lines:
                continue

            # NOTE: Calulate the inverse branch to add it as a single line
            # inverse_line_mask = (pl.col("to_bus") == from_bus.name) & (pl.col("from_bus") == to_bus.name)
            # if not (branch_data.filter(inverse_line_mask).is_empty()):
            #     rating_down = branch_data.filter(inverse_line_mask)["rating"]
            reverse_row = branch_data.filter(
                (pl.col("kind") == branch["kind"])
                & (pl.col("from_bus") == branch["to_bus"])
                & (pl.col("to_bus") == branch["from_bus"])
            ).to_dicts()

            if reverse_row:
                reverse_row_dict = reverse_row[0]
                rating_down = -reverse_row_dict["max_active_power"] * ureg.MW
                reverse_lines.add(
                    (reverse_row_dict["kind"], reverse_row_dict["from_bus"], reverse_row_dict["to_bus"])
                )
            else:
                rating_down = -branch["max_active_power"] * ureg.MW

            losses = branch["losses"] if branch["losses"] else 0
            self.system.add_component(
                self._create_model_instance(
                    MonitoredLine,
                    category=branch["kind"],
                    name=branch_name,
                    from_bus=from_bus,
                    to_bus=to_bus,
                    rating_up=branch["max_active_power"] * ureg.MW,
                    rating_down=rating_down,
                    losses=losses * ureg.percent,
                    ext=ext,
                ),
            )

    def _construct_tx_interfaces(self):
        logger.info("Creating transmission interfaces objects.")
        interface_lines = self.system.get_components(
            MonitoredLine, filter_func=lambda x: x.from_bus.load_zone.name != x.to_bus.load_zone.name
        )
        interfaces = defaultdict(dict)  # Holder of interfaces
        tx_interface_map = self._create_model_instance(TransmissionInterfaceMap, name="transmission_map")
        for line in interface_lines:
            zone_from = line.from_bus.load_zone.name
            zone_to = line.to_bus.load_zone.name
            zone_pair = tuple(sorted((zone_from, zone_to)))
            zone_pair_name = f"{zone_pair[0]}_{zone_pair[1]}"
            if (zone_from, zone_to) == zone_pair:
                positive_flow, negative_flow = line.rating_up.magnitude, line.rating_down.magnitude
            else:
                positive_flow, negative_flow = -line.rating_down.magnitude, -line.rating_up.magnitude

            if "positive_flow" not in interfaces[zone_pair] or "negative_flow" not in interfaces[zone_pair]:
                interfaces[zone_pair]["positive_flow"] = positive_flow
                interfaces[zone_pair]["negative_flow"] = negative_flow
            else:
                interfaces[zone_pair]["positive_flow"] += positive_flow
                interfaces[zone_pair]["negative_flow"] += negative_flow

            tx_interface_map.mapping[zone_pair_name].append(line.label)

        for interface in interfaces:
            interface_name = f"{interface[0]}_{interface[1]}"
            interface_values = interfaces[interface]

            # The inferface transmission flow is defined as the sum of the rating of the lines.
            # In the case that the rate up/down does not match, we take the max and set it in both directions.
            max_power_flow = max(interface_values["positive_flow"], abs(interface_values["negative_flow"]))

            # Ramp multiplier defines the MW/min ratio for the interface
            ramp_multiplier = self.reeds_config.defaults["interface_max_ramp_up_multiplier"]
            self.system.add_component(
                self._create_model_instance(
                    TransmissionInterface,
                    name=interface_name,
                    active_power_flow_limits=MinMax(-max_power_flow, max_power_flow),
                    direction_mapping={},  # TBD
                    ext={
                        "ramp_up": max_power_flow * ramp_multiplier * ureg.Unit("MW/min"),
                        "ramp_down": max_power_flow * ramp_multiplier * ureg.Unit("MW/min"),
                    },
                )
            )
        self.system.add_component(tx_interface_map)

    def _construct_emissions(self) -> None:
        """Construct emission objects."""
        logger.info("Creating emission objects")
        emit_rates = self.get_data("emission_rates")

        emit_rates = emit_rates.with_columns(
            pl.concat_str([pl.col("tech"), pl.col("tech_vintage"), pl.col("region")], separator="_").alias(
                "generator_name"
            ),
            pl.concat_str(
                [
                    pl.col("tech"),
                    pl.col("tech_vintage"),
                    pl.col("region"),
                    pl.col("emission_type"),
                ],
                separator="_",
            ).alias("name"),
        )

        for generator in self.system.get_components(
            Generator, filter_func=lambda x: x.name in emit_rates["generator_name"]
        ):
            generator_emission = emit_rates.filter(pl.col("generator_name") == generator.name)
            for row in generator_emission.iter_rows(named=True):
                row["rate"] = EmissionRate(row["rate"], "kg/MWh")
                row["emission_type"] = get_enum_from_string(row["emission_type"], EmissionType)
                emission_model = self._create_model_instance(Emission, **row)
                self.system.add_component(emission_model)

    def _construct_generators(self) -> None:  # noqa: C901
        """Construct generators objects."""
        logger.info("Creating generator objects.")
        capacity_data = self.get_data("online_capacity")
        generator_fuel = self.get_data("fuels")

        # Fuel price requires two input files
        fuel_price_input = self.get_data("fuel_price")
        bfuel_price_output = (
            self.get_data("bfuel_price").with_columns(fuel=pl.lit("biomass")).join(generator_fuel, on="fuel")
        ).select(pl.exclude("fuel"))
        fuel_price = pl.concat([fuel_price_input, bfuel_price_output], how="diagonal")

        heat_rate = self.get_data("heat_rate")
        cost_vom = self.get_data("cost_vom")
        forced_outages = self.get_data("forced_outages")
        planned_outages = self.get_data("planned_outages")
        storage_duration = self.get_data("storage_duration")
        storage_eff = self.get_data("storage_eff")
        category_map = self.reeds_config.defaults.get("tech_categories", None)

        # NOTE: Temp unit definition. This should be read from the mapping file?
        unit_definition = {
            "fuel_price": "usd/MMBtu",
            "heat_rate": "MMBtu/MWh",
            "planned_outage_rate": "",
            "forced_outage_rate": "",
            "vom_price": "usd/MWh",
            "active_power": "MW",
            "pump_efficiency": "",
            "charge_efficiency": "",
            "discharge_efficiency": "",
            "pump_load": "MW",
            "storage_capacity": "MWh",
            "storage_duration": "h",
            "initial_energy": "%",
            "initial_volume": "MWh",
        }

        # Combine all the generator dataset in a single frame
        gen_data = pl_left_multi_join(
            capacity_data,
            generator_fuel,
            fuel_price,
            heat_rate,
            cost_vom,
            storage_duration,
            storage_eff,
            forced_outages,
            planned_outages,
        )

        # NOTE: Populate fuel_price information for technologies that use bio_fuel
        gen_data = gen_data.with_columns(
            category=pl.col("tech").map_elements(
                lambda row: match_category(row, category_map), return_dtype=pl.String
            )
        )

        # Adding extra columns
        gen_data = gen_data.with_columns(
            [
                (pl.col("active_power") * pl.col("storage_duration")).alias("storage_capacity"),
                (
                    pl.col("active_power")
                    * pl.col("storage_duration")
                    / self.reeds_config.defaults[
                        "initial_volume_divisor"
                    ]  # Storage start at half of its capacity
                ).alias("initial_volume"),
                (pl.lit(100) / self.reeds_config.defaults["initial_volume_divisor"]).alias(
                    "initial_energy"
                ),  # 50%
                (pl.col("active_power")).alias("pump_load"),
                (pl.col("charge_efficiency")).alias("pump_efficiency"),
                (
                    pl.when(pl.col("tech").is_in(self.reeds_config.defaults["commit_technologies"]))
                    .then(1)
                    .otherwise(None)
                ).alias("must_run"),
            ]
        )

        # NOTE: ReEDS exports some technologies that have a online_capacity of zero.
        # Once it is fixed on main we can remove this line over here.
        gen_data = gen_data.filter(pl.col("active_power") > 0)

        # Check that we mapped all categories
        categories = gen_data["category"].unique()
        for category in categories:
            device_map = self.device_map.get(category, "")
            if getattr(R2X_MODELS, device_map, None) is None:
                logger.warning(
                    "Could not parse category {}. Check that the type map is including this category.",
                    category,
                )
        non_cf_generators = gen_data.filter(
            ~pl.col("category").is_in(self.reeds_config.defaults["vre_categories"])
        )
        cf_generators = gen_data.filter(
            pl.col("category").is_in(self.reeds_config.defaults["vre_categories"])
        )
        cf_generators = self._aggregate_renewable_generators(cf_generators)

        combined_data = pl.concat([non_cf_generators, cf_generators], how="align")

        for row in combined_data.iter_rows(named=True):
            category = row["category"]

            if category in self.excluded_categories:
                msg = "`{}` in excluded categories. Skipping it."
                logger.debug(msg, category)
                continue

            device_map = self.device_map.get(category, "")
            if getattr(R2X_MODELS, device_map, None) is None:
                msg = "Could not find device model for `{}`. Skipping it."
                logger.warning(msg, category)
                continue

            gen_model = getattr(R2X_MODELS, device_map)
            for key, value in row.items():
                if key in unit_definition:
                    if value:
                        row[key] = value * ureg.Unit(unit_definition[key])
            # NOTE: We can uncomment this if we define the units onf the REEDS mapping.
            #     if key in self.config.fmap:
            #         units = self.config.fmap[key].get("units", "")
            #         if value is not None:
            #             row[key] = ureg.Quantity(value, units)

            # EXTRACT THIS NAME CONVENTION TO THE CONFIGURATION
            match gen_model.__name__:
                case "RenewableDispatch" | "RenewableNonDispatch":
                    name = row["tech"] + "_" + row["region"]
                case _:
                    name = row["tech"] + "_" + row["tech_vintage"] + "_" + row["region"]

            row["name"] = name

            if not (fuel_pm := self.tech_to_fuel_pm.get(row["category"])) and not self.skip_validation:
                msg = (
                    f"Could not find a fuel and prime mover map for `{row['category']}`."
                    " Check `reeds_input_config.json`"
                )
                raise ParserError(msg)

            row["prime_mover_type"] = (
                get_enum_from_string(fuel_pm["type"], PrimeMoversType) if fuel_pm.get("type") else None
            )
            row["fuel"] = get_enum_from_string(fuel_pm["fuel"], ThermalFuels) if fuel_pm["fuel"] else None

            bus = self.system.get_component(ACBus, name=row["region"])
            row["bus"] = bus
            bus_load_zone = bus.load_zone
            assert bus_load_zone is not None

            # Add reserves/services to generator if they are not excluded
            if row["tech"] not in self.reeds_config.defaults["excluded_reserve_techs"]:
                row["services"] = list(
                    self.system.get_components(
                        Reserve,
                        filter_func=lambda x: x.region.name == bus_load_zone.name,
                    )
                )
                reserve_map = self.system.get_component(ReserveMap, name="reserve_map")
                for reserve_type in row["services"]:
                    reserve_map.mapping[reserve_type.name].append(row["name"])

            # Add operational cost data
            # ReEDS model all the thermal generators assuming an average heat rate
            vom_price = row.get("vom_price", None) or 0.0
            if isinstance(vom_price, Quantity):
                vom_price = vom_price.magnitude
            fuel_price = row.get("fuel_price", None) or 0.0
            if isinstance(fuel_price, Quantity):
                fuel_price = fuel_price.magnitude
            if issubclass(gen_model, RenewableGen):
                row["operation_cost"] = None
            if issubclass(gen_model, ThermalGen):
                if heat_rate := row.get("heat_rate"):
                    if isinstance(heat_rate, Quantity):
                        heat_rate = heat_rate.magnitude
                    heat_rate_curve = AverageRateCurve(
                        function_data=LinearFunctionData(
                            proportional_term=heat_rate,
                            constant_term=0,
                        ),
                        initial_input=heat_rate,
                    )
                    fuel_curve = FuelCurve(
                        value_curve=heat_rate_curve,
                        vom_cost=LinearCurve(vom_price),
                        fuel_cost=fuel_price,
                        power_units=UnitSystem.NATURAL_UNITS,
                    )
                    row["operation_cost"] = ThermalGenerationCost(
                        variable=fuel_curve,
                    )
            if issubclass(gen_model, HydroGen):
                row["operation_cost"] = HydroGenerationCost(
                    variable=CostCurve(
                        value_curve=LinearCurve(vom_price),
                        power_units=UnitSystem.NATURAL_UNITS,
                    )
                )

            row["must_run"] = 1 if row["tech"] in self.reeds_config.defaults["commit_technologies"] else 0

            # NOTE: If there is a point when ReEDs enforces minimum capacity for a technology here is where we
            # will need to change it.
            row["active_power_limits"] = MinMax(min=0, max=row["active_power"].magnitude)

            row["ext"] = {}
            row["ext"] = {
                "tech": row["tech"],
                "reeds_tech": row["tech"],
                "reeds_vintage": row["tech_vintage"],
            }
            self.system.add_component(self._create_model_instance(gen_model, **row))

    def _construct_load(self):
        logger.info("Adding load time series.")

        bus_data = self.get_data("hierarchy")
        load_df = self.get_data("load").collect()
        start = datetime(year=self.weather_year, month=1, day=1)
        resolution = timedelta(hours=1)

        # Calculate starting index for the weather year
        if len(load_df) > 8760:
            end_idx = 8760 * (self.weather_year - BASE_WEATHER_YEAR + 1)  # +1 to be inclusive.
        else:
            end_idx = 8760
        for _, bus_data in enumerate(bus_data.iter_rows(named=True)):
            bus_name = bus_data["region"]
            bus = self.system.get_component(ACBus, name=bus_name)
            ts = SingleTimeSeries.from_array(
                data=ActivePower(load_df[bus_name][end_idx - 8760 : end_idx].to_numpy(), "MW"),
                variable_name="max_active_power",
                initial_time=start,
                resolution=resolution,
            )
            user_dict = {"solve_year": self.reeds_config.weather_year}
            max_load = np.max(ts.data)
            load = self._create_model_instance(
                PowerLoad, name=f"{bus.name}", bus=bus, max_active_power=max_load
            )
            self.system.add_component(load)
            self.system.add_time_series(ts, load, **user_dict)

    def _construct_cf_time_series(self):
        logger.info("Adding cf time series")
        if not self.weather_year:
            raise AttributeError("Missing weather year from the configuration class.")

        cf_data = self.get_data("cf").collect()
        cf_adjustment = self.get_data("cf_adjustment")
        # NOTE: We take the median of  the seasonal adjustment since we
        # aggregate the generators by technology vintage
        cf_adjustment = cf_adjustment.group_by("tech").agg(pl.col("cf_adj").median())
        ilr = self.get_data("ilr")
        ilr = dict(
            ilr.group_by("tech").agg(pl.col("ilr").sum()).iter_rows()
        )  # Dict is more useful here than series
        start = datetime(year=self.weather_year, month=1, day=1)
        resolution = timedelta(hours=1)

        # Calculate starting index for the weather year starting
        if len(cf_data) > 8760:
            end_idx = 8760 * (self.weather_year - BASE_WEATHER_YEAR + 1)  # +1 to be inclusive.
        else:
            end_idx = 8760

        counter = 0
        # NOTE: At some point, I would like to create a single time series per
        # BA instead of attaching one per generator. We would need to invert
        # the order of the loop and just use that to attach it to the different
        for generator in self.system.get_components(RenewableDispatch, RenewableNonDispatch):
            profile_name = generator.name  # .rsplit("_", 1)[0]
            if "|" in cf_data.columns[1]:
                profile_name = "|".join(profile_name.rsplit("_", 1))
            if profile_name not in cf_data.columns:
                msg = (
                    f"{generator.__class__.__name__}:{generator.name} do not "
                    "have a corresponding time series. Consider changing the model to `RenewableGen`"
                )
                logger.warning(msg)
                continue

            cf_adj = cf_adjustment.filter(pl.col("tech") == generator.ext["reeds_tech"])["cf_adj"]
            ilr_value = ilr.get(generator.ext["reeds_tech"], 1)
            rating_profile = (
                generator.active_power
                * ilr_value
                * cf_adj
                * cf_data[profile_name][end_idx - 8760 : end_idx].to_numpy()
            )
            ts = SingleTimeSeries.from_array(
                data=rating_profile,
                variable_name="max_active_power",
                initial_time=start,
                resolution=resolution,
            )
            user_dict = {"solve_year": self.weather_year}
            self.system.add_time_series(ts, generator, **user_dict)
            counter += 1
        logger.debug("Added {} time series objects", counter)

    def _construct_reserve_provision(self):
        # NOTE: We need to re-think this chunk of code. The code is bad.
        # Provision is just based on wind/solar and load for the given region.
        logger.debug("Creating reserve provision")

        # We assume that they all start at the same time, so they have the same
        # resolution of the generator time series
        start = datetime(year=self.weather_year, month=1, day=1)
        resolution = timedelta(hours=1)
        for reserve in self.system.get_components(Reserve):
            region = reserve.region
            provision_objects = {}
            provision_objects["solar"] = [
                component
                for component in self.system.get_components(
                    RenewableDispatch,
                    filter_func=lambda x: x.bus.load_zone.name == region.name
                    and (
                        x.prime_mover_type == PrimeMoversType.PV or x.prime_mover_type == PrimeMoversType.RTPV
                    ),
                )
                if self.system.has_time_series(component)
            ]
            provision_objects["wind"] = [
                component
                for component in self.system.get_components(
                    RenewableDispatch,
                    filter_func=lambda x: x.bus.load_zone.name == region.name
                    and (
                        x.prime_mover_type == PrimeMoversType.WT or x.prime_mover_type == PrimeMoversType.WS
                    ),
                )
                if self.system.has_time_series(component)
            ]
            provision_objects["load"] = [
                component
                for component in self.system.get_components(
                    PowerLoad, filter_func=lambda x: x.bus.load_zone.name == region.name
                )
                if self.system.has_time_series(component)
            ]
            solar_names = list(map(attrgetter("name"), provision_objects["solar"]))
            wind_names = list(map(attrgetter("name"), provision_objects["wind"]))
            load_names = list(map(attrgetter("name"), provision_objects["load"]))
            solar_reserves = list(
                map(
                    getattr,
                    map(
                        getattr,
                        map(self.system.get_time_series, provision_objects["solar"]),
                        repeat("data"),
                    ),
                    repeat("magnitude"),
                )
            )
            solar_capacity = list(
                map(
                    lambda component: self.system.get_component_by_label(
                        component.label
                    ).active_power.magnitude,
                    provision_objects["solar"],
                )
            )
            wind_reserves = list(
                map(
                    getattr,
                    map(
                        getattr,
                        map(self.system.get_time_series, provision_objects["wind"]),
                        repeat("data"),
                    ),
                    repeat("magnitude"),
                )
            )
            load_reserves = list(
                map(
                    getattr,
                    map(
                        getattr,
                        map(self.system.get_time_series, provision_objects["load"]),
                        repeat("data"),
                    ),
                    repeat("magnitude"),
                )
            )
            wind_provision = (
                pa.Table.from_arrays(wind_reserves, names=wind_names)
                .to_pandas()
                .sum(axis=1)
                .mul(self.reeds_config.defaults["wind_reserves"].get(reserve.reserve_type.name, 1))
            )
            solar_provision = (
                pa.Table.from_arrays(solar_reserves, names=solar_names)
                .to_pandas()
                .sum(axis=1)
                .apply(lambda x: 1 if x != 0 else 0)
                .mul(sum(solar_capacity))
                .mul(self.reeds_config.defaults["solar_reserves"].get(reserve.reserve_type.name, 1))
            )
            load_provision = (
                pa.Table.from_arrays(load_reserves, names=load_names)
                .to_pandas()
                .sum(axis=1)
                .mul(self.reeds_config.defaults["load_reserves"].get(reserve.reserve_type.name, 1))
            )
            total_provision = load_provision.add(solar_provision, fill_value=0).add(
                wind_provision, fill_value=0
            )
            if total_provision.empty:
                msg = (
                    f"Reserve provision for {reserve=} is zero."
                    "Check that renewable devices contribute to the reserve"
                )
                logger.warning(msg)
            self.system.add_time_series(
                SingleTimeSeries.from_array(
                    total_provision.to_numpy(),
                    variable_name="requirement",
                    initial_time=start,
                    resolution=resolution,
                ),
                reserve,
            )
            # Add total provision as requirement
            setattr(reserve, "max_requirement", total_provision.sum())

    def _construct_hydro_budgets(self) -> None:
        """Hydro budgets in ReEDS."""
        logger.debug("Adding hydro budgets.")
        month_hrs = read_csv("month_hrs.csv").collect()
        month_map = self.reeds_config.defaults["month_map"]

        hydro_cf = self.get_data("hydro_cf")
        hydro_cf = hydro_cf.with_columns(
            month=pl.col("month").map_elements(lambda row: month_map.get(row, row), return_dtype=pl.String)
        )
        month_hrs = month_hrs.rename({"szn": "season"})
        hydro_data = pl_left_multi_join(
            hydro_cf,
            month_hrs,
        )
        month_of_day = np.array(
            [dt.astype("datetime64[M]").astype(int) % 12 + 1 for dt in self.daily_time_index]
        )
        initial_time = datetime(self.weather_year, 1, 1)
        for generator in self.system.get_components(HydroDispatch):
            # NOTE: Canadian imports need another file for the ratings, but we process it as
            # HydroEnergyReservoir since it is the way ReEDS model it.
            if generator.category == "can-imports":
                continue
            tech = generator.ext["reeds_tech"]
            generator_bus = generator.bus
            assert generator_bus
            region = generator_bus.name
            hydro_ratings = hydro_data.filter((pl.col("tech") == tech) & (pl.col("region") == region))

            hourly_time_series = np.zeros(len(month_of_day), dtype=float)

            for row in hydro_ratings.iter_rows(named=True):
                month = row["month"]
                if isinstance(month, str):
                    month = int(month.removeprefix("M"))

                month_max_budget = (
                    generator.active_power * Percentage(row["hydro_cf"], "") * Time(row["hrs"], "h")
                )
                daily_max_budget = month_max_budget / (row["hrs"] / 24)
                hourly_time_series[month_of_day == month] = daily_max_budget.magnitude

            ts = SingleTimeSeries.from_array(
                Energy(hourly_time_series / 1e3, "GWh"),
                "hydro_budget",
                initial_time=initial_time,
                resolution=timedelta(days=1),
            )
            self.system.add_time_series(ts, generator)

        return None

    def _construct_hydro_rating_profiles(self) -> None:
        logger.debug("Adding hydro rating profiles.")
        month_hrs = read_csv("month_hrs.csv").collect()
        month_map = self.reeds_config.defaults["month_map"]

        hydro_cf = self.get_data("hydro_cf")
        hydro_cf = hydro_cf.with_columns(
            month=pl.col("month").map_elements(lambda row: month_map.get(row, row), return_dtype=pl.String)
        )
        month_hrs = month_hrs.rename({"szn": "season"})
        hydro_data = pl_left_multi_join(
            hydro_cf,
            month_hrs,
        )
        month_of_hour = np.array(
            [dt.astype("datetime64[M]").astype(int) % 12 + 1 for dt in self.hourly_time_index]
        )
        initial_time = datetime(self.weather_year, 1, 1)
        for generator in self.system.get_components(HydroEnergyReservoir):
            tech = generator.ext["reeds_tech"]
            generator_bus = generator.bus
            assert generator_bus is not None
            region = generator_bus.name
            generator.inflow = 0.0
            generator.initial_storage = generator.initial_energy
            generator.storage_capacity = Energy(0.0, "MWh")
            generator.storage_target = Energy(0.0, "MWh")

            hourly_time_series = np.zeros(len(month_of_hour), dtype=float)
            hydro_ratings = hydro_data.filter((pl.col("tech") == tech) & (pl.col("region") == region))
            for row in hydro_ratings.iter_rows(named=True):
                month = row["month"]
                if isinstance(month, str):
                    month = int(month.removeprefix("M"))
                month_indices = month_of_hour == month
                rating = generator.active_power * Percentage(row["hydro_cf"], "")
                hourly_time_series[month_indices] = rating.magnitude
            ts = SingleTimeSeries.from_array(
                ActivePower(hourly_time_series, "MW"),
                "max_active_power",
                initial_time=initial_time,
                resolution=timedelta(hours=1),
            )
            self.system.add_time_series(ts, generator)
        return None

    def _construct_hybrid_systems(self):
        """Create hybrid storage units and add them to the system."""
        hybrids = list(
            self.system.get_components(RenewableDispatch, filter_func=lambda x: "pvb" in x.ext["tech"])
        )

        for device in hybrids:
            hybrid_values = device.model_dump()
            hybrid_name = hybrid_values.pop("name")

            ext_dict = device.ext

            # The only two required fields to create the new units are
            # `active_power` and the `bus`. We add some validation on both to
            # guarantee that the parser is correctly adding those into the ext
            # dictionary.
            bus: ACBus = device.bus
            msg = "Bus object not found for hybtid system. Check ingestion of the model inputs"
            assert bus is not None and isinstance(bus, Bus), msg

            # Create storage asset for hybrid.
            storage_unit_fields = {
                key: value for key, value in ext_dict.items() if key in GenericBattery.model_fields
            }
            storage_unit_fields["prime_mover_type"] = PrimeMoversType.BA

            # If at some point we change the power of the storage it should be here
            storage_unit = self._create_model_instance(
                GenericBattery,
                name=f"{hybrid_name}",
                active_power=device.active_power,  # Assume same power for the battery
                category="pvb-storage",
                bus=bus,
                ext=ext_dict,
                **storage_unit_fields,
            )
            self.system.add_component(storage_unit)

            # Updating name of the renewable component
            new_device = self.system.copy_component(device, name=f"{device.name}", attach=True)
            ts = self.system.get_time_series(device)
            self.system.add_time_series(ts, new_device)
            self.system.remove_component(device)
            hybrid_construct = self._create_model_instance(
                HybridSystem, name=f"{hybrid_name}", renewable_unit=new_device, storage_unit=storage_unit
            )
            self.system.add_component(hybrid_construct)

    def _aggregate_renewable_generators(self, data: pl.DataFrame) -> pl.DataFrame:
        return data.group_by(["tech", "region"]).agg(
            [pl.col("active_power").sum(), pl.exclude("active_power").first()]
        )

    def _create_model_instance(self, model_class, **kwargs):
        return create_model_instance(model_class, skip_validation=self.skip_validation, **kwargs)



"""Utilities for the configuration."""

import inspect
from loguru import logger

from .config_models import BaseModelConfig, MODEL_CONFIGS, Models, ExporterModels, ParserModels

from .utils import read_json, read_fmap


def get_input_defaults(model_enum: Models) -> dict:
    """Return configuration dicitonary based on the input model."""
    defaults_dict = read_json("r2x/defaults/config.json")
    plugins_dict = read_json("r2x/defaults/plugins_config.json")

    defaults_dict = defaults_dict | plugins_dict
    match model_enum:
        case Models.INFRASYS:
            logger.debug("Returning infrasys defaults")
        case Models.REEDS:
            defaults_dict = defaults_dict | read_json("r2x/defaults/reeds_input.json")
            logger.debug("Returning reeds defaults")
        case Models.SIENNA:
            defaults_dict = defaults_dict | read_json("r2x/defaults/sienna_config.json")
            logger.debug("Returning sienna defaults")
        case Models.PLEXOS:
            defaults_dict = defaults_dict | read_json("r2x/defaults/plexos_input.json")
            logger.debug("Returning input_model {} defaults", model_enum)
        case _:
            msg = (
                f"Unsupported input model: {model_enum}. "
                f"Supported models: {[str(model) for model in ParserModels]}"
            )
            raise ValueError(msg)
    return defaults_dict


def get_output_defaults(model_enum: Models) -> dict:
    """Return configuration dicitonary based on the output model."""
    match model_enum:
        case Models.INFRASYS:
            # NOTE: Here we will add any infrasys configuration if we need in the future.
            defaults_dict = None
            pass
        case Models.PLEXOS:
            defaults_dict = (
                read_json("r2x/defaults/plexos_output.json")
                | read_json("r2x/defaults/plexos_simulation_objects.json")
                | read_json("r2x/defaults/plexos_horizons.json")
                | read_json("r2x/defaults/plexos_models.json")
            )
            logger.debug("Returning output_model {} defaults", model_enum)
        case Models.SIENNA:
            defaults_dict = read_json("r2x/defaults/sienna_config.json")
            logger.debug("Returning sienna defaults")
        case _:
            msg = (
                f"Unsupported input model: {model_enum}. "
                f"Supported models: {[str(model) for model in ExporterModels]}"
            )
            raise ValueError(msg)

    if not defaults_dict:
        return {}

    return defaults_dict


def get_input_model_fmap(model_enum: Models) -> dict:
    """Return input model file mape based on the model_name."""
    match model_enum:
        case Models.INFRASYS:
            fmap = {}
        case Models.REEDS:
            fmap = read_fmap("r2x/defaults/reeds_us_mapping.json")
        case Models.SIENNA:
            fmap = read_fmap("r2x/defaults/sienna_mapping.json")
        case Models.PLEXOS:
            fmap = read_fmap("r2x/defaults/plexos_mapping.json")
        case _:
            raise ValueError(f"Input model {model_enum=} not valid")
    return fmap


def get_model_config_class(model_enum: Models, **kwargs) -> BaseModelConfig:
    """Get the appropriate model configuration from the model name."""
    if model_enum not in MODEL_CONFIGS:
        msg = f"Unsupported model: {model_enum}. Supported models: {list(MODEL_CONFIGS.keys())}"
        raise ValueError(msg)

    model_config = MODEL_CONFIGS[model_enum]

    # Extract fields for the given config
    cls_fields = {field for field in inspect.signature(model_config).parameters}
    model_kwargs = {key: value for key, value in kwargs.items() if key in cls_fields}
    model_config_class = model_config(**model_kwargs)

    # Add fmap and defaults
    model_config_class.fmap = get_input_model_fmap(model_enum)
    return model_config_class



import json

import pytest
import yaml

from r2x.utils import haskey, read_user_dict


@pytest.mark.utils
@pytest.mark.parametrize(
    "base_dict, path, expected",
    [
        ({"a": {"b": {"c": 1}}}, ["a", "b", "c"], True),  # valid path
        ({"a": {"b": {"c": 1}}}, ["a", "b"], True),  # valid partial path
        ({"a": {"b": {"c": 1}}}, ["a", "x"], False),  # invalid key in path
        ({"a": {"b": {"c": 1}}}, ["a", "b", "d"], False),  # invalid path at last level
        ({}, ["a"], False),  # empty dictionary
        ({"a": {"b": {"c": 1}}}, ["a", "b", "c", "d"], False),  # path too long
    ],
)
def test_haskey(base_dict, path, expected):
    assert haskey(base_dict, path) == expected


def test_read_user_dict(tmp_path):
    user_dict = '{"fmap": true}'
    parsed_dict = read_user_dict(user_dict)

    assert isinstance(parsed_dict, dict)
    assert parsed_dict.get("fmap", None) is not None

    user_dict = '{"fmap": true : 2}'
    with pytest.raises(ValueError):
        _ = read_user_dict(user_dict)

    user_dict = '[{"fmap": true}]'
    with pytest.raises(ValueError):
        _ = read_user_dict(user_dict)

    sample_data = {"name": "test", "fmap": True, "input_model": "plexos"}

    yaml_file = tmp_path / "data.yaml"
    with yaml_file.open("w") as f:
        yaml.dump(sample_data, f)

    assert yaml_file.exists()
    parsed_dict = read_user_dict(str(yaml_file))
    assert isinstance(parsed_dict, dict)
    assert parsed_dict.get("fmap", None) is not None
    assert parsed_dict["input_model"] == "plexos"

    json_file = tmp_path / "data.json"
    with json_file.open("w") as f:
        json.dump(sample_data, f, indent=4)

    assert json_file.exists()
    parsed_dict = read_user_dict(str(json_file))
    assert isinstance(parsed_dict, dict)
    assert parsed_dict.get("fmap", None) is not None
    assert parsed_dict["input_model"] == "plexos"

    yaml_file = tmp_path / "missing_file.yaml"
    with pytest.raises(FileNotFoundError):
        _ = read_user_dict(str(yaml_file))



"""Plexos parser class."""

import importlib
from argparse import ArgumentParser
from collections.abc import Sequence
from datetime import datetime, timedelta
from importlib.resources import files
from pathlib import Path, PureWindowsPath
from typing import Any

import numpy as np
import polars as pl
from infrasys.cost_curves import CostCurve, FuelCurve, UnitSystem
from infrasys.exceptions import ISNotStored
from infrasys.function_data import (
    LinearFunctionData,
    QuadraticFunctionData,
)
from infrasys.time_series_models import SingleTimeSeries
from infrasys.value_curves import AverageRateCurve, InputOutputCurve, LinearCurve
from loguru import logger
from pint import Quantity
from plexosdb import PlexosSQLite
from plexosdb.enums import ClassEnum, CollectionEnum

from r2x.api import System
from r2x.config_models import PlexosConfig
from r2x.enums import ACBusTypes, PrimeMoversType, ReserveDirection, ReserveType, ThermalFuels
from r2x.exceptions import ModelError, ParserError
from r2x.models import (
    ACBus,
    Generator,
    GenericBattery,
    HydroDispatch,
    LoadZone,
    MonitoredLine,
    RenewableGen,
    Reserve,
    ReserveMap,
    ThermalGen,
    TransmissionInterface,
    TransmissionInterfaceMap,
)
from r2x.models.branch import Transformer2W
from r2x.models.core import MinMax
from r2x.models.costs import HydroGenerationCost, RenewableGenerationCost, ThermalGenerationCost
from r2x.models.load import PowerLoad
from r2x.units import ureg
from r2x.utils import get_enum_from_string, get_pint_unit, validate_string

from .handler import PCMParser, csv_handler
from .parser_helpers import (
    construct_pwl_from_quadtratic,
    field_filter,
    prepare_ext_field,
    reconcile_timeseries,
)
from .plexos_utils import (
    DATAFILE_COLUMNS,
    PLEXOS_ACTION_MAP,
    filter_property_dates,
    find_xml,
    get_column_enum,
    parse_data_file,
    time_slice_handler,
)
from .polars_helpers import pl_filter_year

models = importlib.import_module("r2x.models")

R2X_MODELS = importlib.import_module("r2x.models")
BASE_WEATHER_YEAR = 2007
XML_FILE_KEY = "xml_file"
DEFAULT_QUERY_COLUMNS_SCHEMA = {  # NOTE: Order matters
    "membership_id": pl.Int64,
    "parent_object_id": pl.Int32,
    "parent_object_name": pl.String,
    "parent_class_name": pl.String,
    "child_class_name": pl.String,
    "category": pl.String,
    "object_id": pl.Int32,
    "name": pl.String,
    "property_name": pl.String,
    "property_unit": pl.String,
    "property_value": pl.Float32,
    "band": pl.Int32,
    "date_from": pl.String,
    "date_to": pl.String,
    "memo": pl.String,
    "scenario_category": pl.String,
    "scenario": pl.String,
    "action": pl.String,
    "data_file_tag": pl.String,
    "data_file": pl.String,
    "variable_tag": pl.String,
    "timeslice_tag": pl.String,
    "timeslice": pl.String,
    "timeslice_value": pl.Float32,
    "data_text": pl.String,
}
DEFAULT_PROPERTY_COLUMNS = [
    "band",
    "property_name",
    "property_value",
    "property_unit",
    "data_file",
    "variable",
    "action",
    # "variable_tag",
    # "variable_default",
    # "timeslice",
    # "timeslice_value",
]
DEFAULT_INDEX = ["object_id", "name", "category"]
SIMPLE_QUERY_COLUMNS_SCHEMA = {
    "parent_class_name": pl.String,
    "child_class_name": pl.String,
    "parent_object_id": pl.Int32,
    "object_id": pl.Int32,
    "name": pl.String,
    "category": pl.String,
    "property_name": pl.String,
    "property_unit": pl.String,
    "property_value": pl.Float64,
    "band": pl.Int32,
    "date_from": pl.String,
    "date_to": pl.String,
    "text": pl.String,
    "text_class_name": pl.String,
    "tag_timeslice": pl.String,
    "tag_timeslice_object_id": pl.Int32,
    "tag_datafile": pl.String,
    "tag_datafile_object_id": pl.Int32,
    "tag_variable": pl.String,
    "tag_variable_object_id": pl.Int32,
    "action": pl.String,
    "scenario": pl.String,
}


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--model",
        required=False,
        help="Plexos model to translate",
    )
    return parser


class PlexosParser(PCMParser):
    """Plexos parser class."""

    def __init__(self, *args, xml_file: str | None = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert self.config.run_folder
        assert self.config.input_config
        assert isinstance(self.config.input_config, PlexosConfig)  # Only take Plexos configurations
        self.run_folder = Path(self.config.run_folder)
        self.input_config = self.config.input_config
        self.system = System(name=self.config.name, auto_add_composed_components=True)
        self.property_map = self.input_config.defaults["plexos_input_property_map"] or {}
        self.device_map = self.input_config.defaults["plexos_device_map"] or {}
        self.fuel_map = self.input_config.defaults["plexos_fuel_map"] or {}
        self.category_map = self.input_config.defaults["plexos_category_map"] or {}
        self.device_match_string = self.input_config.defaults["device_name_inference_map"] or {}
        self.generator_models = self.input_config.defaults["generator_models"] or {}
        self.year = self.input_config.model_year
        assert self.year
        assert isinstance(self.year, int)

        # TODO(pesap): Rename exceptions to include R2X
        # https://github.com/NREL/R2X/issues/5
        # R2X needs at least one of this maps defined to correctly work.
        one_required = ["fuel_map", "device_map", "device_match_string", "category_map"]
        if all(getattr(self, one_req, {}) == {} for one_req in one_required):
            msg = f"At least one of {', or '.join(one_required)} is required to initialize PlexosParser"
            raise ParserError(msg)

        # Populate databse from XML file.
        # If xml file is not specified, check user_dict["fmap"]["xml_file"] or use
        # only xml file in project directory
        if xml_file is None:
            xml_file = self.input_config.fmap.get("xml_file", {}).get("fname", None)
            xml_file = xml_file or str(find_xml(self.run_folder))

        xml_file = str(self.run_folder / xml_file)

        self.db = PlexosSQLite(xml_fname=xml_file)

        # Extract scenario data
        model_name = getattr(self.input_config, "model_name", None) or self.input_config.fmap.get(
            "xml_file", {}
        ).get("model_name", None)
        if model_name is None:
            model_name = self._select_model_name()
        logger.info("Parsing plexos model={}", model_name)
        self._process_scenarios(model_name=model_name)

        # date from is in days since 1900, convert to year
        if not self.year:
            logger.debug("Getting ST year from model horizon")
            date_from = self._collect_horizon_data(model_name=model_name).get("Date From")
            if date_from is not None:
                self.year = int((date_from / 365.25) + 1900)

        self.hourly_time_index = pl.datetime_range(
            datetime(self.year, 1, 1), datetime(self.year + 1, 1, 1), interval="1h", eager=True, closed="left"
        ).to_frame("datetime")

        return

    def build_system(self) -> System:
        """Create infrasys system."""
        logger.info("Building infrasys system using {}", self.__class__.__name__)

        # If we decide to change the engine for handling the data we can do it here.
        object_data = self._plexos_table_data()
        self.plexos_data = self._polarize_data(object_data=object_data)

        # Construct the network
        self._construct_load_zones()
        self._construct_buses()
        self._construct_branches()
        self._construct_transformers()
        self._construct_interfaces()
        self._construct_reserves()

        # Generators
        self._construct_generators()
        self._add_buses_to_generators()
        self._add_generator_reserves()

        # Batteries
        self._construct_batteries()
        self._add_buses_to_batteries()
        self._add_battery_reserves()

        self._construct_load_profiles()
        return self.system

    def _collect_horizon_data(self, model_name: str) -> dict[str, float]:
        """Collect horizon data (Date From/To) from Plexos database."""
        horizon_query = f"""
        SELECT
            atr.name AS attribute_name,
            COALESCE(attr_data.value, atr.default_value) AS attr_val
        FROM
            t_object
        LEFT JOIN t_class AS class ON
            t_object.class_id == class.class_id
        LEFT JOIN t_attribute AS atr ON
            t_object.class_id  == atr.class_id
        LEFT JOIN t_membership AS tm ON
            t_object.object_id  == tm.child_object_id
        LEFT JOIN t_class AS parent_class ON
            tm.parent_class_id == parent_class.class_id
        LEFT JOIN t_object AS to2 ON
            tm.parent_object_id == to2.object_id
        LEFT JOIN t_attribute_data attr_data ON
            attr_data.attribute_id == atr.attribute_id AND t_object.object_id == attr_data.object_id
        WHERE
            class.name == '{ClassEnum.Horizon.value}'
            AND parent_class.name == '{ClassEnum.Model.value}'
            AND to2.name == '{model_name}'
        """
        horizon_data = self.db.query(horizon_query)
        horizon_map = {key: value for key, value in horizon_data}
        return horizon_map

    def _get_fuel_prices(self):
        logger.debug("Creating fuel representation")
        system_fuels = (pl.col("child_class_name") == ClassEnum.Fuel.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )

        fuels = self._get_model_data(system_fuels)
        fuels.write_csv("fuels.csv")
        fuel_prices = {}
        for fuel_name, fuel_data in fuels.group_by("name"):
            fuel_name = fuel_name[0]
            property_records = fuel_data.to_dicts()

            for property in property_records:
                property.update({"property_unit": "$/MMBtu"})

            mapped_records, multi_band_records = self._parse_property_data(property_records)
            if multi_band_records:
                logger.warning("Some properties have multiple bands.")
            mapped_records["name"] = fuel_name
            fuel_prices[fuel_name] = mapped_records.get("Price", 0)
        return fuel_prices

    def _construct_load_zones(self, default_model=LoadZone) -> None:
        """Create LoadZone representation.

        Plexos can define load at multiple levels, but for balancing the load,
        we assume that it happens at the region level, which is a typical way
        of doing it.
        """
        logger.info("Creating load zone representation")
        system_regions = (pl.col("child_class_name") == ClassEnum.Region.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        regions = self._get_model_data(system_regions)

        region_pivot = regions.pivot(
            index=DEFAULT_INDEX,
            on="property_name",
            values="property_value",
            aggregate_function="first",
        )
        for region in region_pivot.iter_rows(named=True):
            valid_fields, ext_data = field_filter(region, default_model.model_fields)
            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))
        return

    def _construct_buses(self, default_model=ACBus) -> None:
        logger.info("Creating buses representation")
        system_buses = (pl.col("child_class_name") == ClassEnum.Node.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        region_buses = (pl.col("child_class_name") == ClassEnum.Region.value) & (
            pl.col("parent_class_name") == ClassEnum.Node.value
        )
        system_buses = self._get_model_data(system_buses)
        buses_region = self._get_model_data(region_buses)
        for idx, (bus_name, bus_data) in enumerate(system_buses.group_by("name")):
            bus_name = bus_name[0]
            logger.trace("Parsing bus = {}", bus_name)

            property_records = bus_data.to_dicts()
            mapped_records, _ = self._parse_property_data(property_records)
            mapped_records["name"] = bus_name
            # mapped_records, _ = self._parse_property_data(property_records)

            valid_fields, ext_data = field_filter(mapped_records, default_model.model_fields)

            # Get region from buses region memberships
            region_name = buses_region.filter(pl.col("parent_object_id") == bus_data["object_id"].unique())[
                "name"
            ].item()

            valid_fields["load_zone"] = self.system.get_component(LoadZone, name=region_name)

            # We parse all the buses as PV unless in the model someone specify a bus is a slack.
            # Plexos defines the True value of a Slack bus assigning it -1. Possible values are only 0
            # (False), -1 (True).
            valid_fields["bustype"] = ACBusTypes.PV
            if mapped_records.get("Is Slack Bus") == -1:
                valid_fields["bustype"] = ACBusTypes.SLACK

            valid_fields["base_voltage"] = (
                230.0 if not valid_fields.get("base_voltage") else valid_fields["base_voltage"]
            )

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            bus = default_model(number=idx + 1, **valid_fields)
            self.system.add_component(bus)

            if max_active_power := mapped_records.pop("max_active_power", False):
                max_load = (
                    np.nanmax(max_active_power.data)
                    if isinstance(max_active_power, SingleTimeSeries)
                    else max_active_power
                )
                # We only add the load if it is bigger than zero.
                if max_load > 0:
                    load = PowerLoad(name=f"{bus_name}", bus=bus, max_active_power=max_load)
                    self.system.add_component(load)
                    ts_dict = {"solve_year": self.year}
                    if isinstance(max_active_power, SingleTimeSeries):
                        self.system.add_time_series(max_active_power, load, **ts_dict)

            ts_fields = {k: v for k, v in mapped_records.items() if isinstance(v, SingleTimeSeries)}
            if ts_fields:
                generator = self.system.get_component_by_label(f"{default_model.__name__}.{bus_name}")
                ts_dict = {"solve_year": self.year}
                for ts_name, ts in ts_fields.items():
                    ts.variable_name = ts_name
                    self.system.add_time_series(ts, generator, **ts_dict)
        return

    def _construct_reserves(self, default_model=Reserve):
        logger.info("Creating reserve representation")

        system_reserves = self._get_model_data(
            (pl.col("child_class_name") == ClassEnum.Reserve.name)
            & (pl.col("parent_class_name") == ClassEnum.System.name)
        )

        for reserve_name, reserve_data in system_reserves.group_by("name"):
            reserve_name = reserve_name[0]
            logger.trace("Parsing reserve = {}", reserve_name)
            property_records = reserve_data.to_dicts()
            mapped_records, _ = self._parse_property_data(property_records)
            mapped_records["name"] = reserve_name
            reserve_type = validate_string(mapped_records.pop("Type", "default"))
            plexos_reserve_map = self.input_config.defaults["reserve_types"].get(
                str(reserve_type), self.input_config.defaults["reserve_types"]["default"]
            )  # Pass string so we do not need to convert the json mapping.
            mapped_records["reserve_type"] = ReserveType[plexos_reserve_map["type"]]
            mapped_records["direction"] = ReserveDirection[plexos_reserve_map["direction"]]

            # Service model uses all floats
            mapped_records["max_requirement"] = (
                mapped_records.pop("max_requirement").magnitude
                if mapped_records.get("max_requirement")
                else None
            )
            mapped_records["vors"] = mapped_records.pop("vors").magnitude
            mapped_records["duration"] = mapped_records["duration"].magnitude
            if mapped_records["time_frame"].units == "second":
                mapped_records["time_frame"] = mapped_records["time_frame"].magnitude / 60
            else:
                mapped_records["time_frame"] = mapped_records["time_frame"].magnitude

            valid_fields, ext_data = field_filter(mapped_records, default_model.model_fields)
            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))

        reserve_map = ReserveMap(name="contributing_generators")
        self.system.add_component(reserve_map)
        return

    def _construct_branches(self, default_model=MonitoredLine):
        logger.info("Creating lines")
        system_lines = (pl.col("child_class_name") == ClassEnum.Line.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        system_lines = self._get_model_data(system_lines)
        lines_pivot = system_lines.pivot(
            index=DEFAULT_INDEX,
            on="property_name",
            values="property_value",
            aggregate_function="first",
        )
        if lines_pivot.is_empty():
            logger.warning("No line objects found on the system.")
            return

        lines_pivot_memberships = self.db.get_memberships(
            *lines_pivot["name"].to_list(), object_class=ClassEnum.Line
        )
        for line in lines_pivot.iter_rows(named=True):
            line_properties_mapped = {self.property_map.get(key, key): value for key, value in line.items()}
            line_properties_mapped["rating"] = line_properties_mapped.get("max_power_flow", 0.0)
            line_properties_mapped["rating_up"] = line_properties_mapped.pop("max_power_flow", 0.0)
            line_properties_mapped["rating_down"] = line_properties_mapped.pop("min_power_flow", 0.0)

            if line_properties_mapped["rating"] is None:
                logger.warning("Skipping disabled line {}", line)
                continue

            valid_fields, ext_data = field_filter(line_properties_mapped, default_model.model_fields)

            from_bus_name = next(
                membership
                for membership in lines_pivot_memberships
                if (
                    membership[2] == line["name"]
                    and membership[5] == ClassEnum.Line.name
                    and membership[6].replace(" ", "") == CollectionEnum.NodeFrom.name
                )
            )[3]
            from_bus = self.system.get_component(ACBus, from_bus_name)
            to_bus_name = next(
                membership
                for membership in lines_pivot_memberships
                if (
                    membership[2] == line["name"]
                    and membership[5] == ClassEnum.Line.name
                    and membership[6].replace(" ", "") == CollectionEnum.NodeTo.name
                )
            )[3]
            to_bus = self.system.get_component(ACBus, to_bus_name)
            valid_fields["from_bus"] = from_bus
            valid_fields["to_bus"] = to_bus

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))
        return

    def _construct_transformers(self, default_model=Transformer2W):
        logger.info("Creating transformers")
        system_transformers = (pl.col("child_class_name") == ClassEnum.Transformer.value) & (
            pl.col("parent_class_name") == ClassEnum.System.value
        )
        system_transformers = self._get_model_data(system_transformers)
        transformer_pivot = system_transformers.pivot(
            index=DEFAULT_INDEX,
            on="property_name",
            values="property_value",
            aggregate_function="first",
        )
        if transformer_pivot.is_empty():
            logger.warning("No transformer objects found on the system.")
            return

        lines_pivot_memberships = self.db.get_memberships(
            *transformer_pivot["name"].to_list(), object_class=ClassEnum.Transformer
        )
        for transformer in transformer_pivot.iter_rows(named=True):
            transformer_properties_mapped = {
                self.property_map.get(key, key): value for key, value in transformer.items()
            }
            transformer_properties_mapped["rating"] = transformer_properties_mapped.get(
                "max_active_power", 0.0
            )
            transformer_properties_mapped["rating_up"] = transformer_properties_mapped.pop(
                "max_power_flow", 0.0
            )
            transformer_properties_mapped["rating_down"] = transformer_properties_mapped.pop(
                "min_power_flow", 0.0
            )

            if transformer_properties_mapped["rating"] is None:
                logger.warning("Skipping disabled transformer {}", transformer)
                continue

            valid_fields, ext_data = field_filter(transformer_properties_mapped, default_model.model_fields)

            from_bus_name = next(
                membership
                for membership in lines_pivot_memberships
                if (
                    membership[2] == transformer["name"]
                    and membership[5] == ClassEnum.Transformer.name
                    and membership[6].replace(" ", "") == CollectionEnum.NodeFrom.name
                )
            )[3]
            from_bus = self.system.get_component(ACBus, from_bus_name)
            to_bus_name = next(
                membership
                for membership in lines_pivot_memberships
                if (
                    membership[2] == transformer["name"]
                    and membership[5] == ClassEnum.Transformer.name
                    and membership[6].replace(" ", "") == CollectionEnum.NodeTo.name
                )
            )[3]
            to_bus = self.system.get_component(ACBus, to_bus_name)
            valid_fields["from_bus"] = from_bus
            valid_fields["to_bus"] = to_bus

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))
        return

    def _infer_model_type(self, generator_name):
        inference_mapper = self.device_match_string
        generator_name_lower = generator_name.lower()
        for key, device_info in inference_mapper.items():
            if key in generator_name_lower:
                return device_info
        return None

    def _get_fuel_pmtype(
        self, generator_name, fuel_name: str | None = None, category: str | None = None
    ) -> dict[str, str]:
        """Return fuel prime mover combo based on priority.

        This function will return first a match using the generator name, then
        by fuel map and if not it will try to infer it using the generator name

        Returns
        -------
        dict
            Either an empty dictionary or a dictionary with fuel and prime mover type.
        """
        match_fuel_pmtype = (
            self.category_map.get(category)
            or self.device_map.get(generator_name)
            or self.fuel_map.get(fuel_name)
            or self._infer_model_type(generator_name)
        )
        if not match_fuel_pmtype:
            return {}

        msg = (
            "Mapping returned a different data structure. "
            f"Returned {type(match_fuel_pmtype)} and expected type dict."
        )
        assert isinstance(match_fuel_pmtype, dict), msg
        return match_fuel_pmtype

    def _get_model_type(self, fuel_pmtype) -> str:
        """Get model type for given pair of fuel and prime mover type.

        Notes
        -----
        We return an empty string if not match found since it is used for the
        getattr function. If you pass a None it does not work.

        Returns
        -------
        str
            Match found or Empty string if not match is found.
        """
        assert len(self.generator_models) < 1000, "Change the data structure. Fool."
        if fuel_pmtype is None:
            return ""
        for model, conditions in self.generator_models.items():
            for cond in conditions:
                if (cond["fuel"] == fuel_pmtype["fuel"] or cond["fuel"] is None) and (
                    cond["type"] == fuel_pmtype["type"] or cond["type"] is None
                ):
                    return model
        return ""

    def _construct_generators(self):  # noqa: C901
        """Create Plexos generator objects."""
        logger.info("Creating generator objects")

        # Filter only generator objects that belong to the system
        # NOTE: Polars 1.1.0 can not make a comparisson between a Enum and a string column. If we convert the
        # Enum to string it works, but we might be able to change this in a near future if polars supports it.
        system_generators_filter = (pl.col("child_class_name") == str(ClassEnum.Generator)) & (
            pl.col("parent_class_name") == str(ClassEnum.System)
        )
        system_generators = self._get_model_data(system_generators_filter)
        if self.config.feature_flags.get("plexos-csv", None):
            system_generators.write_csv("generators.csv")

        # NOTE: The best way to identify the type of generator on Plexos is by reading the fuel
        fuel_query = f"""
        SELECT
            parent_object.name AS parent_object_name,
            child_object.name AS fuel_name
        FROM
            t_membership AS mem
            LEFT JOIN t_object AS child_object ON mem.child_object_id = child_object.object_id
            LEFT JOIN t_object AS parent_object ON mem.parent_object_id = parent_object.object_id
            LEFT JOIN t_class AS child_class ON child_object.class_id = child_class.class_id
            LEFT JOIN t_class AS parent_cls ON parent_object.class_id = parent_cls.class_id
        WHERE
            child_class.name = '{ClassEnum.Fuel}'
            AND parent_cls.name = '{ClassEnum.Generator}'
        """
        generator_fuel = self.db.query(fuel_query)
        generator_fuel_map = {key: value for key, value in generator_fuel}

        fuel_prices = self._get_fuel_prices()

        # Iterate over properties to create generator object
        for generator_name, generator_data in system_generators.group_by("name"):
            generator_name = generator_name[0]

            category = generator_data["category"].unique()
            if len(category) > 1:
                msg = "Generator has more then one category. Check the dataset"
                logger.debug(msg)
            category = category[0]

            fuel_name = generator_fuel_map.get(generator_name)
            logger.trace("Parsing generator = {} with fuel type = {}", generator_name, fuel_name)

            fuel_pmtype = self._get_fuel_pmtype(generator_name, fuel_name=fuel_name, category=category)

            if not fuel_pmtype:
                msg = "Fuel mapping not found for {} with fuel_type={}"
                logger.warning(msg, generator_name, fuel_name)
                continue

            # We assume that if we find the fuel, there is a model_map
            model_map = self._get_model_type(fuel_pmtype)
            model_map = getattr(R2X_MODELS, model_map)

            # property_records = generator_data[DEFAULT_PROPERTY_COLUMNS].to_dicts()
            property_records = generator_data.to_dicts()

            mapped_records, multi_band_records = self._parse_property_data(property_records)
            mapped_records["name"] = generator_name

            # if multi_band_records:
            #     pass

            # Get prime mover enum
            mapped_records["prime_mover_type"] = fuel_pmtype["type"]
            mapped_records["prime_mover_type"] = PrimeMoversType[mapped_records["prime_mover_type"]]
            mapped_records["fuel"] = (
                get_enum_from_string(fuel_pmtype["fuel"], ThermalFuels) if fuel_pmtype.get("fuel") else None
            )

            # Pumped Storage generators are not required to have Max Capacity property
            if "max_active_power" not in mapped_records and "pump_load" in mapped_records:
                mapped_records["rating"] = mapped_records["pump_load"]

            mapped_records = self._set_unit_capacity(mapped_records)
            if mapped_records is None:
                logger.trace("Skipping disabled generator {}", generator_name)
                # When unit availability is not set, we skip the generator
                continue

            required_fields = {
                key: value for key, value in model_map.model_fields.items() if value.is_required()
            }
            if not all(key in mapped_records for key in required_fields):
                missing_fields = [key for key in required_fields if key not in mapped_records]
                logger.warning(
                    "Skipping generator {}. Missing Required Fields: {}", generator_name, missing_fields
                )
                continue

            if not mapped_records.get("fuel_price"):
                mapped_records["fuel_price"] = fuel_prices.get(generator_fuel_map.get(generator_name), 0.0)
            mapped_records = self._construct_operating_costs(mapped_records, generator_name, model_map)

            valid_fields, ext_data = field_filter(mapped_records, model_map.model_fields)

            ts_fields = {k: v for k, v in mapped_records.items() if isinstance(v, SingleTimeSeries)}

            valid_fields.update(
                {
                    ts_name: np.mean(ts.data)
                    for ts_name, ts in ts_fields.items()
                    if ts_name in valid_fields.keys()
                }
            )

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(model_map(**valid_fields))
            generator = self.system.get_component_by_label(f"{model_map.__name__}.{generator_name}")

            if ts_fields:
                generator = self.system.get_component_by_label(f"{model_map.__name__}.{generator_name}")
                ts_dict = {"solve_year": self.year}
                for ts_name, ts in ts_fields.items():
                    ts.variable_name = ts_name
                    self.system.add_time_series(ts, generator, **ts_dict)
        return

    def _add_buses_to_generators(self):
        # Add buses to generators
        generators = [generator["name"] for generator in self.system.to_records(Generator)]
        generator_memberships = self.db.get_memberships(
            *generators,
            object_class=ClassEnum.Generator,
            collection=CollectionEnum.Nodes,
        )
        for generator in self.system.get_components(Generator):
            buses = [membership for membership in generator_memberships if membership[2] == generator.name]
            if buses:
                for bus in buses:
                    try:
                        bus_object = self.system.get_component(ACBus, name=bus[3])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for generator:{} since reserve {} is not stored",
                            generator.name,
                            buses[3],
                        )
                        continue
                    generator.bus = bus_object
        return

    def _add_generator_reserves(self):
        reserve_map = self.system.get_component(ReserveMap, name="contributing_generators")
        generators = [generator["name"] for generator in self.system.to_records(Generator)]
        generator_memberships = self.db.get_memberships(
            *generators,
            object_class=ClassEnum.Generator,
            parent_class=ClassEnum.Reserve,
            collection=CollectionEnum.Generators,
        )
        for generator in self.system.get_components(Generator):
            reserves = [membership for membership in generator_memberships if membership[3] == generator.name]
            if reserves:
                # NOTE: This would get replaced if we have a method on infrasys
                # that check if something exists on the system
                for reserve in reserves:
                    try:
                        reserve_object = self.system.get_component(Reserve, name=reserve[2])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for generator:{} since reserve {} is not stored",
                            generator.name,
                            reserve[2],
                        )
                        continue
                    reserve_map.mapping[reserve_object.name].append(generator.name)
        return

    def _construct_batteries(self):
        logger.info("Creating battery objects")
        system_batteries = self._get_model_data(
            (pl.col("child_class_name") == ClassEnum.Battery.name)
            & (pl.col("parent_class_name") == ClassEnum.System.name)
        )

        required_fields = {
            key: value for key, value in GenericBattery.model_fields.items() if value.is_required()
        }

        for battery_name, battery_data in system_batteries.group_by("name"):
            battery_name = battery_name[0]
            logger.trace("Parsing battery = {}", battery_name)

            property_records = battery_data.to_dicts()

            mapped_records, _ = self._parse_property_data(property_records)
            mapped_records["name"] = battery_name
            mapped_records["prime_mover_type"] = PrimeMoversType.BA

            mapped_records = self._set_unit_capacity(mapped_records)
            if mapped_records is None:
                continue

            valid_fields, ext_data = field_filter(mapped_records, GenericBattery.model_fields)

            if not all(key in valid_fields for key in required_fields):
                missing_fields = [key for key in required_fields if key not in valid_fields]
                logger.warning(
                    "Skipping battery {}. Missing required fields: {}", battery_name, missing_fields
                )
                continue

            if mapped_records["storage_capacity"] == 0:
                logger.warning("Skipping battery {} since it has zero capacity", battery_name)
                continue

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(GenericBattery(**valid_fields))
        return

    def _add_buses_to_batteries(self):
        batteries = [battery["name"] for battery in self.system.to_records(GenericBattery)]
        if not batteries:
            msg = "No battery objects found on the system. Skipping adding membership to buses"
            logger.warning(msg)
            return
        generator_memberships = self.db.get_memberships(
            *batteries,
            object_class=ClassEnum.Battery,
            collection=CollectionEnum.Nodes,
        )
        for component in self.system.get_components(GenericBattery):
            buses = [membership for membership in generator_memberships if membership[2] == component.name]
            if buses:
                for bus in buses:
                    try:
                        bus_object = self.system.get_component(ACBus, name=bus[3])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for battery:{} since bus {} is not stored",
                            component.name,
                            buses[3],
                        )
                        continue
                    component.bus = bus_object
        return

    def _add_battery_reserves(self):
        reserve_map = self.system.get_component(ReserveMap, name="contributing_generators")
        batteries = [battery["name"] for battery in self.system.to_records(GenericBattery)]
        if not batteries:
            msg = "No battery objects found on the system. Skipping adding reserve memberships"
            logger.warning(msg)
            return
        generator_memberships = self.db.get_memberships(
            *batteries,
            object_class=ClassEnum.Battery,
            parent_class=ClassEnum.Reserve,
            collection=CollectionEnum.Batteries,
        )

        for battery in self.system.get_components(GenericBattery):
            reserves = [membership for membership in generator_memberships if membership[3] == battery.name]
            if reserves:
                # NOTE: This would get replaced if we have a method on infrasys
                # that check if something exists on the system
                for reserve in reserves:
                    try:
                        reserve_object = self.system.get_component(Reserve, name=reserve[2])
                    except ISNotStored:
                        logger.warning(
                            "Skipping membership for generator:{} since reserve {} is not stored",
                            battery.name,
                            reserve[2],
                        )
                        continue
                    reserve_map.mapping[reserve_object.name].append(battery.name)
        return

    def _construct_interfaces(self, default_model=TransmissionInterface):
        """Construct Transmission Interface and Transmission Interface Map."""
        logger.info("Creating transmission interfaces")
        system_interfaces_mask = (pl.col("child_class_name") == ClassEnum.Interface.name) & (
            pl.col("parent_class_name") == ClassEnum.System.name
        )
        system_interfaces = self._get_model_data(system_interfaces_mask)
        interfaces = system_interfaces.pivot(
            index=DEFAULT_INDEX,
            on="property_name",
            values="property_value",
            aggregate_function="first",
        )

        interface_property_map = {
            v: k
            for k, v in self.input_config.defaults["plexos_input_property_map"].items()
            if k in default_model.model_fields
        }

        tx_interface_map = TransmissionInterfaceMap(name="transmission_map")
        for interface in interfaces.iter_rows(named=True):
            mapped_interface = {
                interface_property_map.get(key, key): value for key, value in interface.items()
            }
            valid_fields, ext_data = field_filter(mapped_interface, default_model.model_fields)

            # Check that the interface has all the required fields of the model.
            required_fields = {
                key: value for key, value in default_model.model_fields.items() if value.is_required()
            }
            if not all(key in mapped_interface for key in required_fields):
                missing_fields = [key for key in required_fields if key not in mapped_interface]
                logger.warning(
                    "{}:{} missing required fields: {}. Skipping it.",
                    default_model.__name__,
                    interface["name"],
                    missing_fields,
                )
                continue

            valid_fields = prepare_ext_field(valid_fields, ext_data)
            self.system.add_component(default_model(**valid_fields))

        # Add lines memberships
        lines = [line["name"] for line in self.system.to_records(MonitoredLine)]
        if not lines:
            return
        lines_memberships = self.db.get_memberships(
            *lines,
            object_class=ClassEnum.Line,
            parent_class=ClassEnum.Interface,
            collection=CollectionEnum.Lines,
        )
        for line in self.system.get_components(MonitoredLine):
            interface = next(
                (membership for membership in lines_memberships if membership[3] == line.name), None
            )
            if interface:
                # NOTE: This would get replaced if we have a method on infrasys
                # that check if something exists on the system
                try:
                    interface_object = self.system.get_component(TransmissionInterface, name=interface[2])
                except ISNotStored:
                    logger.warning(
                        "Skipping membership for line:{} since interface {} is not stored",
                        line.name,
                        interface[2],
                    )
                    continue
                tx_interface_map.mapping[interface_object.name].append(line.label)
        self.system.add_component(tx_interface_map)
        return

    def _construct_value_curves(self, record, generator_name):  # noqa: C901
        """Construct value curves for generators."""
        if not any("Heat Rate" in key or "heat_rate" in key for key in record.keys()):
            return None
        vc = None
        heat_rate_avg = record.get("heat_rate", None)
        heat_rate_base = record.get("Heat Rate Base", None)
        heat_rate_incr = record.get("Heat Rate Incr", None)
        heat_rate_incr2 = record.get("Heat Rate Incr2", None)

        if any(isinstance(val, SingleTimeSeries) for val in [heat_rate_avg, heat_rate_base, heat_rate_incr]):
            logger.warning(
                "Time-varying heat-rates not implemented for generator={}. Using median value instead",
                generator_name,
            )

        heat_rate_avg = (
            Quantity(
                np.median(heat_rate_avg.data),
                units=heat_rate_avg.data.units,
            )
            if isinstance(heat_rate_avg, SingleTimeSeries)
            else heat_rate_avg
        )
        heat_rate_base = (
            Quantity(
                np.median(heat_rate_base.data),
                units=heat_rate_base.data.units,
            )
            if isinstance(heat_rate_base, SingleTimeSeries)
            else heat_rate_base
        )
        heat_rate_incr = (
            Quantity(
                np.median(heat_rate_incr.data),
                units=heat_rate_incr.data.units,
            )
            if isinstance(heat_rate_incr, SingleTimeSeries)
            else heat_rate_incr
        )

        if heat_rate_incr and heat_rate_incr.units == "british_thermal_unit / kilowatt_hour":
            heat_rate_incr = Quantity(heat_rate_incr.magnitude * 1e-3, "british_thermal_unit / watt_hour")

        if heat_rate_avg and heat_rate_avg.units == "british_thermal_unit / kilowatt_hour":
            heat_rate_avg = Quantity(heat_rate_avg.magnitude * 1e-3, "british_thermal_unit / watt_hour")

        if heat_rate_avg:
            fn = LinearFunctionData(proportional_term=heat_rate_avg.magnitude, constant_term=0)
            vc = AverageRateCurve(
                function_data=fn,
                initial_input=heat_rate_avg.magnitude,
            )
        elif heat_rate_incr2 and "** 2" in str(heat_rate_incr2.units):
            fn = QuadraticFunctionData(
                quadratic_term=heat_rate_incr2.magnitude,
                proportional_term=heat_rate_incr.magnitude,
                constant_term=heat_rate_base.magnitude,
            )
            if self.config.feature_flags.get("quad2pwl", None):
                n_tranches = self.config.feature_flags.get("quad2pwl", None)
                fn = construct_pwl_from_quadtratic(fn, record, n_tranches)
        elif not heat_rate_incr2 and heat_rate_incr:
            fn = LinearFunctionData(
                proportional_term=heat_rate_incr.magnitude, constant_term=heat_rate_base.magnitude
            )
        else:
            logger.warning("Heat Rate type not implemented for generator={}", generator_name)
            fn = None

        if not vc and fn:
            vc = InputOutputCurve(function_data=fn)
        if not vc and fn is None:
            vc = LinearCurve(0)
        return vc

    def _construct_operating_costs(self, mapped_records, generator_name, model_map):
        """Construct operating costs from Value Curves and Operating Costs."""
        vom_cost = mapped_records.get("vom_price", 0.0)
        if isinstance(vom_cost, Quantity):
            vom_cost = vom_cost.magnitude

        if issubclass(model_map, RenewableGen):
            mapped_records["operation_cost"] = RenewableGenerationCost()
        elif issubclass(model_map, ThermalGen):
            heat_rate_curve = self._construct_value_curves(mapped_records, generator_name)
            fuel_cost = mapped_records.get("fuel_price", 0)
            if isinstance(fuel_cost, SingleTimeSeries):
                fuel_cost = np.mean(fuel_cost.data)
            elif isinstance(fuel_cost, Quantity):
                fuel_cost = fuel_cost.magnitude
            if heat_rate_curve:
                cost_curve = FuelCurve(
                    value_curve=heat_rate_curve, fuel_cost=fuel_cost, power_units=UnitSystem.NATURAL_UNITS
                )
            else:
                cost_curve = CostCurve(
                    value_curve=LinearCurve(0),
                    vom_cost=LinearCurve(vom_cost),
                    power_units=UnitSystem.NATURAL_UNITS,
                )
            mapped_records["operation_cost"] = ThermalGenerationCost(
                variable=cost_curve,
                start_up=mapped_records.get("startup_cost", 0),
            )
        elif issubclass(model_map, HydroDispatch):
            mapped_records["operation_cost"] = HydroGenerationCost()
        else:
            logger.warning(
                "Operating Cost not implemented for generator={} model map={}", generator_name, model_map
            )

        mapped_records["vom_price"] = mapped_records.get("vom_price", 0)
        return mapped_records

    def _select_model_name(self):
        # TODO(pesap): Handle exception if no model name found
        # https://github.com/NREL/R2X/issues/10
        query = f"""
        select obj.name
        from t_object as obj
        left join t_class as cls on obj.class_id = cls.class_id
        where cls.name = '{ClassEnum.Model.name}'
        """
        result = self.db.query(query)
        model_names = [row[0] for row in result]

        print("Available models:")
        for idx, name in enumerate(model_names):
            print(f"{idx + 1}. {name}")

        while True:
            try:
                choice = int(input("Select a model by number: "))
                if 1 <= choice <= len(model_names):
                    return model_names[choice - 1]
                else:
                    print(f"Please enter a number between 1 and {len(model_names)}.")
            except ValueError:
                print("Invalid input. Please enter a number.")

    def _process_scenarios(self, model_name: str | None = None) -> None:
        """Create a SQLite representation of the XML."""
        if model_name is None:
            msg = "Required model name not found. Parser requires a model to parse from the Plexos database"
            raise ModelError(msg)
        # self.db = PlexosSQLite(xml_fname=xml_fpath)

        logger.trace("Getting object_id for model={}", model_name)
        model_id_query = """
        SELECT
            object_id
        FROM
            t_object
        LEFT JOIN
            t_class on t_class.class_id = t_object.class_id
        WHERE t_object.name = ? and t_class.name = ?
        """
        model_id = self.db.query(
            model_id_query,
            params=(model_name, ClassEnum.Model),
        )
        if len(model_id) > 1:
            msg = f"Multiple models with the same {model_name} returned. Check database or spelling."
            logger.debug(model_id)
            raise ModelError(msg)
        if not model_id:
            msg = f"Model `{model_name}` not found on the XML. Check spelling of the `model_name`."
            raise ParserError(msg)
        self.model_id = model_id[0][0]  # Unpacking tuple [(model_id,)]

        # NOTE: When doing performance updates this query could get some love.
        valid_scenarios = self.db.query(
            "select obj.name from t_membership mem "
            "left join t_object as obj on obj.object_id = mem.child_object_id "
            "left join t_class as cls on cls.class_id = obj.class_id "
            f"where mem.parent_object_id = {self.model_id} and cls.name = '{ClassEnum.Scenario}'"
        )
        valid_scenarios_query = """
        SELECT
            t_object.name
        FROM
            t_membership as membership
        LEFT JOIN
            t_object on t_object.object_id = membership.child_object_id
        LEFT JOIN
            t_class on t_object.class_id = t_class.class_id
        WHERE
            membership.parent_object_id = ? and
            t_class.name = ?
        """
        valid_scenarios = self.db.query(
            valid_scenarios_query,
            params=(self.model_id, ClassEnum.Scenario),
        )
        if not valid_scenarios:
            msg = f"{model_name=} does not have any scenario attached to it."
            logger.warning(msg)
            return
        assert valid_scenarios
        self.scenarios = [scenario[0] for scenario in valid_scenarios]  # Flatten list of tuples
        return None

    def _set_unit_capacity(self, record):  # noqa: C901
        """Set availability and active power limit TS for generators.

        Plexos does not have a property for base_power (MW) instead it defines it as Max Capacity. In some
        cases, users will derate this Max Capacity using the Rating property which overrules the Max
        Capacity and it becomes the new max_active_power of the generator. This property can be defined as a
        single value or a value that changes by pattern or a time series. If defined as a time series, we
        take the get the maximum active power limits.

        If the Rating property exist and is not defined for the entire time series, Plexos assumes that the
        times not covered the Rating = 0.

        Plexos allows to define a Rating Factor that is expressed as percentage of of the Max Capacity or
        Rating if it is defined. If this property is active, we update the active_power_limits to match the
        new rating.

        Plexos models multiple generators by changing the units attributes (that gets translated to
        available here). If the `available` key is 0, it means that the unit is deactivated from the model or
        retired. If it is 1, it means a single unit that is active.

        Notes
        -----
        At this point of the code, all the properties should be either a single value, or a `SingleTimeSeries`
        """
        if not (availability := record.get("available")):
            logger.warning("Unit `{}` is not activated on the model.", record["name"])
            return
        record["active_power_limits"] = self._get_active_power_limits(record)

        # Set availability, rating, storage_capacity as multiplier of availability/'units'
        if (
            record.get("storage_capacity") is not None
            and not isinstance(record.get("available"), SingleTimeSeries)
            and availability > 1
        ):
            record["storage_capacity"] *= availability

        # If we have Rating, we need to re-adjust the max_active_power and the active_power limits
        # Rating can be either a single number or a time series.
        if rating := record.get("rating"):
            if isinstance(rating, SingleTimeSeries):
                rating = rating.data
                if rating.units == "percent":
                    rating = rating.to("")
                # We override the max active power to the max rating
                record["max_active_power"] = np.nanmax(rating)
                record["active_power_limits_max"] = np.nanmax(rating)
            else:
                record["max_active_power"] = rating * availability
                record["active_power_limits_max"] = rating * availability

        # If we have rating factor we apply it to the max_active power
        # NOTE: There could be cases where the Rating Factor applies to the rating instead of the Max
        # Capacity. For those cases, we will need to re-adjust this function.
        if rating_factor := record.get("Rating Factor"):
            if isinstance(rating_factor, SingleTimeSeries):
                rating_factor_data = rating_factor.data
                if rating_factor_data.units == "percent":
                    rating_factor_data = rating_factor_data.to("")
                if not (max_active_power := record.get("max_active_power")):
                    # Order of the operation matters
                    record["max_active_power"] = availability * record["base_power"] * rating_factor_data
                else:
                    record["max_active_power"] = availability * max_active_power.data * rating_factor_data
                if not isinstance(record["max_active_power"], SingleTimeSeries):
                    record["max_active_power"] = SingleTimeSeries.from_array(
                        record["max_active_power"].data,
                        variable_name="max_active_power",
                        initial_time=rating_factor.initial_time,
                        resolution=rating_factor.resolution,
                    )
            else:
                record["max_active_power"] = rating_factor * record["base_power"] * availability

        # Since we use available as if the unit is active or not, we need to set it back to 1
        record["available"] = 1

        return record

    def _get_active_power_limits(self, record) -> MinMax:
        # assert record["base_power"] is not None
        if active_power_min := record.get("min_rated_capacity"):
            if isinstance(active_power_min, SingleTimeSeries):
                active_power_min = np.nanmin(active_power_min.data)
        active_power_limits_min = active_power_min or 0.0

        if active_power_max := record.get("max_active_power"):
            if isinstance(active_power_max, SingleTimeSeries):
                active_power_max = np.nanmax(active_power_max.data)
        active_power_limits_max = active_power_max or record["base_power"]
        return MinMax(active_power_limits_min, active_power_limits_max)

    def _plexos_table_data(self) -> list[tuple]:
        sql_query = files("plexosdb.queries").joinpath("simple_object_query.sql").read_text()
        object_data = self.db.query(sql_query)
        return object_data

    def _polarize_data(self, object_data: list[tuple]) -> pl.DataFrame:
        data = pl.from_records(object_data, schema=SIMPLE_QUERY_COLUMNS_SCHEMA)

        # Create a lookup map to find nested objects easily
        object_map = data.select(["object_id", "tag_datafile", "tag_datafile_object_id"]).to_dict(
            as_series=False
        )
        self.id_to_name = dict(zip(object_map["object_id"], object_map["tag_datafile"]))
        self.id_to_tag_id = dict(zip(object_map["object_id"], object_map["tag_datafile_object_id"]))
        return data

    def _get_model_data(self, data_filter) -> pl.DataFrame:
        """Filter plexos data for a given class and all scenarios in a model."""
        assert isinstance(self.year, int)
        scenario_specific_data = None
        scenario_filter = None
        if getattr(self, "scenarios", None):
            scenario_filter = pl.col("scenario").is_in(self.scenarios)
            scenario_specific_data = self.plexos_data.filter(data_filter & scenario_filter)
            scenario_specific_data = filter_property_dates(scenario_specific_data, self.year)

        base_case_filter = pl.col("scenario").is_null()
        # Default is to parse data normally if there is not scenario. If scenario exist modify the filter.
        if scenario_specific_data is None:
            system_data = self.plexos_data.filter(data_filter & base_case_filter)
            system_data = filter_property_dates(system_data, self.year)
        else:
            # include both scenario specific and basecase data
            combined_key_base = pl.col("name") + "_" + pl.col("property_name")
            combined_key_scenario = (
                scenario_specific_data["name"] + "_" + scenario_specific_data["property_name"]
            )

            base_case_filter = base_case_filter & (
                ~combined_key_base.is_in(combined_key_scenario) | pl.col("property_name").is_null()
            )
            base_case_data = self.plexos_data.filter(data_filter & base_case_filter)
            base_case_data = filter_property_dates(base_case_data, self.year)

            system_data = pl.concat([scenario_specific_data, base_case_data])

        # If date_from / date_to is specified, override the base_case_value
        rows_to_keep = system_data.filter(pl.col("date_from").is_not_null() | pl.col("date_to").is_not_null())
        rtk_key = rows_to_keep["name"] + "_" + rows_to_keep["property_name"]
        sys_data_key = system_data["name"] + "_" + system_data["property_name"]
        if not rows_to_keep.is_empty():
            # remove if name and property_name are the same
            system_data = system_data.filter(~sys_data_key.is_in(rtk_key))
            system_data = pl.concat([system_data, rows_to_keep])

        return system_data

    def _construct_load_profiles(self):
        logger.info("Creating load profile time series")
        system_regions = (pl.col("child_class_name") == ClassEnum.Region.name) & (
            pl.col("parent_class_name") == ClassEnum.System.name
        )
        regions = self._get_model_data(system_regions)
        for region, region_data in regions.group_by("name"):
            property_records = region_data.to_dicts()
            mapped_records, _ = self._parse_property_data(property_records)

            if max_active_power := mapped_records.get("max_active_power"):
                max_load = (
                    np.nanmax(max_active_power.data)
                    if isinstance(max_active_power, SingleTimeSeries)
                    else max_active_power
                )
            else:
                continue
            bus_region_membership = self.db.get_memberships(
                region[0],
                object_class=ClassEnum.Region,
                parent_class=ClassEnum.Node,
                collection=CollectionEnum.Region,
            )
            for bus in bus_region_membership:
                bus = self.system.get_component(ACBus, name=bus[2])
                load = PowerLoad(name=f"{bus.name}", bus=bus, max_active_power=max_load)
                self.system.add_component(load)
                ts_dict = {"solve_year": self.year}
                if isinstance(max_active_power, SingleTimeSeries):
                    self.system.add_time_series(max_active_power, load, **ts_dict)
        return

    def _data_file_handler(
        self,
        record_name: str,
        property_name: str,
        fpath_str: str,
        variable_name: str | None = None,
        csv_file_encoding: str = "utf8",
    ):
        """Read time varying data from a data file."""
        assert isinstance(self.year, int)

        if encoding := self.config.feature_flags.get("csv_file_encoding"):
            csv_file_encoding = encoding

        # Adjust for Windows type of paths
        if "\\" in fpath_str:
            path = self.run_folder / PureWindowsPath(fpath_str)
        else:
            path = self.run_folder / Path(fpath_str)

        data_file = csv_handler(path, csv_file_encoding=csv_file_encoding)

        column_type = get_column_enum(data_file.columns)
        if column_type is None:
            msg = f"Time series format {data_file.columns=} not yet supported."
            raise NotImplementedError(msg)

        parsed_file = self._parse_data_file(data_file, record_name, property_name, column_type, variable_name)

        if parsed_file.is_empty():
            msg = "Could not find record_name = {} or property_name = {} in fpath = {}. Check data file."
            logger.warning(msg, record_name, property_name, path)
            return

        single_value = self._get_single_value(parsed_file, record_name, property_name, variable_name)
        if single_value is not None:
            return single_value

        if _ := self.config.feature_flags.get("simplify-heat-rate") and property_name == "Heat Rate":
            logger.debug("Simplified time series heat rate for {}", record_name)
            return parsed_file["value"].median()

        # PATCH: It is not common to have a Max Capacity time series. But if it is the case, we just get the
        # median as well
        if property_name == "Max Capacity":
            logger.trace("Simplified Max Capacity time series for {}", record_name)
            return parsed_file["value"].median()

        if "year" not in parsed_file.columns:
            parsed_file = parsed_file.with_columns(year=self.year)

        columns_to_check = self._create_columns_to_check(column_type)

        if not parsed_file.filter(parsed_file.select(columns_to_check).is_duplicated()).is_empty():
            logger.warning("File {} has duplicated rows. Removing duplicates.", path)
            parsed_file = parsed_file.unique(subset=columns_to_check).sort(pl.all())

        # We reconcile the time series data using the hourly time stamp given by the solve year

        parsed_file = reconcile_timeseries(parsed_file, hourly_time_index=self.hourly_time_index)
        assert "value" in parsed_file.columns, (
            f"Error: column value not found on time series file for {record_name}:{property_name}"
        )
        return parsed_file["value"].cast(pl.Float64).to_numpy()

    def _create_columns_to_check(self, column_type: DATAFILE_COLUMNS):
        # NOTE: Some files might have duplicated data. If so, we warn the user and drop the duplicates.
        columns_to_check = [
            column
            for column in column_type.value
            if column in ["name", "pattern", "year", "datetime", "month", "day", "period", "hour"]
        ]
        if column_type == DATAFILE_COLUMNS.TS_YMDH or column_type == DATAFILE_COLUMNS.TS_NMDH:
            columns_to_check.append("hour")
        if column_type == DATAFILE_COLUMNS.TS_NM:
            columns_to_check.append("month")
        return columns_to_check

    def _parse_data_file(
        self,
        data_file,
        record_name: str,
        property_name: str,
        column_type: DATAFILE_COLUMNS,
        variable_name: str | None = None,
    ) -> pl.DataFrame:
        """Parse and filter data based on record and property names."""
        assert isinstance(self.year, int)
        parsed_file = parse_data_file(column_type, data_file)

        if "year" in parsed_file.columns:
            parsed_file = pl_filter_year(parsed_file, year=self.year)

            if parsed_file.is_empty():
                logger.warning("No time series data specified for year filter. Year passed {}", self.year)

        if "name" in parsed_file.columns:
            cols = [col.lower() for col in [record_name, property_name, variable_name] if col]
            parsed_file = parsed_file.filter(pl.col("name").str.to_lowercase().is_in(cols))

        return parsed_file

    def _get_single_value(
        self,
        parsed_file: pl.DataFrame,
        record_name: str,
        property_name: str,
        variable_name: str | None = None,
    ) -> float | None:
        """Return a single value from the parsed file if it matches certain conditions."""
        if not len(parsed_file) == 1:
            return None

        # Return if there is only a value column specified
        if "value" in parsed_file.columns:
            return parsed_file["value"][0]

        if property_name.lower() in parsed_file.columns:
            return parsed_file[property_name.lower()][0]

        names_to_check = [name for name in (record_name, property_name, variable_name) if name]
        if any(name.lower() in parsed_file["name"] for name in names_to_check):
            return parsed_file["value"][0]
        return None

    def _parse_value(self, value: Any, variable_name: str | None = None, unit: str | None = None):
        """Return appropiate value with units if passed."""
        if not isinstance(value, np.ndarray | Sequence):
            return value * ureg.Unit(unit) if unit else value

        assert isinstance(self.year, int)
        assert variable_name
        initial_time = datetime(self.year, 1, 1)
        resolution = timedelta(hours=1)

        return SingleTimeSeries(
            data=ureg.Quantity(value, unit) if unit else value,  # type: ignore
            variable_name=variable_name,
            initial_time=initial_time,
            resolution=resolution,
        )

    def _apply_action(self, action, val_a, val_b):
        val_a_data = val_a.data if isinstance(val_a, SingleTimeSeries) else val_a
        val_b_data = val_b.data if isinstance(val_b, SingleTimeSeries) else val_b

        results = action(val_a_data, val_b_data)
        if isinstance(val_a, SingleTimeSeries):
            val_a.data = results
            return val_a
        if isinstance(val_b, SingleTimeSeries):
            val_b.data = results
            return val_b
        return results

    def _parse_property_data(self, record_data: list[dict[str, Any]]):
        mapped_properties = {}
        property_counts: dict[str, Any] = {}
        multi_band_properties = set()
        timeslice_properties = set()
        unit = None
        mapped_property_name = None
        property_unit_map = {}

        for record in record_data:
            band = record["band"]
            prop_name = record["property_name"]
            prop_value = record["property_value"]
            timeslice = record["tag_timeslice"]
            unit = record["property_unit"].replace("$", "usd") if record["property_unit"] else None
            mapped_property_name = self.property_map.get(prop_name, prop_name)
            unit = get_pint_unit(unit)
            property_unit_map[mapped_property_name] = unit
            value = self._handle_record(record, prop_name, prop_value, unit)

            if mapped_property_name not in property_counts:
                mapped_properties[mapped_property_name] = value
                property_counts[mapped_property_name] = {band}
                property_counts[mapped_property_name] = {
                    "count": 1,
                    "bands": {band},
                    "timeslices": {timeslice},
                    "values": {band: {timeslice: value}},  # Store values by band and timeslice
                }
            else:
                property_counts[mapped_property_name]["count"] += 1
                property_counts[mapped_property_name]["bands"].add(band)
                property_counts[mapped_property_name]["timeslices"].add(timeslice)

                # Capture the value for multi-band/multi-timeslice
                if band not in property_counts[mapped_property_name]["values"]:
                    property_counts[mapped_property_name]["values"][band] = {}
                property_counts[mapped_property_name]["values"][band][timeslice] = value

                # Handle multi-band and multi-timeslice cases
                if len(property_counts[mapped_property_name]["bands"]) > 1:
                    multi_band_properties.add(mapped_property_name)
                if len(property_counts[mapped_property_name]["timeslices"]) > 1:
                    timeslice_properties.add(mapped_property_name)

                mapped_properties[mapped_property_name] = value
        if timeslice_properties:
            for property in timeslice_properties:
                property_with_timeslice = property_counts[property]
                pattern_values = []
                for timeslice in property_with_timeslice["timeslices"]:
                    timeslice_object_id = self.db.get_object_id(timeslice, class_name=ClassEnum.Timeslice)
                    timeslice_data = self._filter_by_object_id(timeslice_object_id)
                    pattern_values.append(
                        {
                            "pattern": timeslice_data["text"][0],
                            "value": property_with_timeslice["values"][1][
                                timeslice
                            ],  # 1 since we only assume single band timeslices
                        }
                    )
                mapped_properties[property] = self._parse_value(
                    time_slice_handler(pattern_values, self.hourly_time_index),
                    property,
                    unit=property_unit_map[property],
                )
        return mapped_properties, multi_band_properties

    def _handle_record(self, record: dict[str, Any], prop_name, prop_value, unit):  # noqa: C901
        """Handle record data.

        Possible cases:
        1. The record has a property value withouth any text, tag_timeslice, tag_variable, tag_datafile

        Possible time series cases:
        1. A Data File is defined for a property as a text
        2. A Data File is defined for a property as a tag with a nested DataFile

        Possible timeslice cases values:
        1. A time slice is defined for a property as a text,
        2. A time slice is defined for a property as a tag with a nested time slice,

        Possible variable cases values. Not that variable needs to have knowdlege of the action
        1. A variable is defined for a property as a tag with a time slice,
        2. A variable is defined for a property as a tag with a data file
        3. A variable is defined for a property as a tag with a nested data file
        4. A variable is defined for a property as a text with a data file
        5. A variable is defined for a property as a text with a nested data file
        """
        action = PLEXOS_ACTION_MAP.get(record["action"], None) if record.get("action") else None
        mapped_property_name = self.property_map.get(prop_name, prop_name)
        match record:
            case {"text": None, "tag_timeslice": None, "tag_datafile": None, "tag_variable": None}:
                logger.trace("Parsing standard property")
                value = self._parse_value(prop_value, unit=unit)
            case {"text": str(), "text_class_name": ClassEnum.DataFile}:
                data_file_value = self._data_file_handler(record["name"], prop_name, record["text"])
                if data_file_value is None:
                    data_file_value = prop_value
                value = self._parse_value(
                    value=data_file_value, variable_name=mapped_property_name, unit=unit
                )
            case {"text": str(), "text_class_name": ClassEnum.Variable}:
                nested_object_id = self.db.get_object_id(record["text"], class_name=ClassEnum.Variable)
                nested_object_data = self._get_nested_object_data(nested_object_id)
                if isinstance(nested_object_data, str):
                    value = (
                        self._data_file_handler(
                            record_name=record["name"],
                            property_name=prop_name,
                            fpath_str=nested_object_data,
                            variable_name=record["text"],
                        )
                        or prop_value
                    )
                else:
                    value = nested_object_data
                if action:
                    value = self._apply_action(action, prop_value, value)
                value = self._parse_value(value, variable_name=mapped_property_name, unit=unit)

            # This case covers when the variable is used to scale a property that is nested on a data file
            case {"tag_datafile": str(), "tag_variable": str()}:
                nested_object_id = self.db.get_object_id(
                    record["tag_variable"], class_name=ClassEnum.Variable
                )
                nested_object_data = self._get_nested_object_data(nested_object_id)
                if isinstance(nested_object_data, str):
                    nested_object_data = self._data_file_handler(
                        record_name=record["name"],
                        property_name=prop_name,
                        fpath_str=nested_object_data,
                        variable_name=record["tag_variable"],
                    )
                record["text"] = self._get_nested_object_data(record["tag_datafile_object_id"])
                data_file_value = self._data_file_handler(
                    record["name"], prop_name, fpath_str=str(record["text"])
                )
                if data_file_value is None:
                    data_file_value = prop_value
                value = self._apply_action(action, data_file_value, nested_object_data)
                value = self._parse_value(value, variable_name=mapped_property_name, unit=unit)
            case {"tag_datafile": str()}:
                record["text"] = self._get_nested_object_data(record["tag_datafile_object_id"])
                data_file_value = self._data_file_handler(
                    record["name"], prop_name, fpath_str=str(record["text"])
                )
                if data_file_value is None:
                    data_file_value = prop_value
                value = self._parse_value(data_file_value, variable_name=mapped_property_name, unit=unit)
            case {"tag_variable": str()}:
                nested_object_id = self.db.get_object_id(
                    record["tag_variable"], class_name=ClassEnum.Variable
                )
                nested_object_data = self._get_nested_object_data(nested_object_id)
                if isinstance(nested_object_data, str):
                    value = self._data_file_handler(
                        record_name=record["name"],
                        property_name=prop_name,
                        fpath_str=nested_object_data,
                        variable_name=record["tag_variable"],
                    )
                else:
                    value = nested_object_data
                if action:
                    value = self._apply_action(action, prop_value, value)
                value = self._parse_value(value, variable_name=mapped_property_name, unit=unit)
            case {"tag_timeslice": str()}:
                value = self._parse_value(prop_value, variable_name=mapped_property_name, unit=unit)
            case _:
                msg = f"Record format class not yet supported. {record=}"
                raise NotImplementedError(msg)

        return value

    def _resolve_object_id(self, object_id: int) -> int:
        """Recursively resolve the object ID by following the tag_object_id when tag_object_name is defined.

        Parameters
        ----------
        object_id : int
            The initial object ID to resolve.

        Returns
        -------
        int
            The resolved object ID that does not have a tag_object_name.
        """
        assert self.id_to_name
        assert self.id_to_tag_id
        while True:
            tag_object_name = self.id_to_name.get(object_id)
            if tag_object_name is None:
                return object_id

            object_id = self.id_to_tag_id[object_id]

    def _filter_by_object_id(self, object_id: int) -> pl.DataFrame:
        """Filter the DataFrame to return rows that match the specified object_id.

        Parameters
        ----------
        object_id : int
            The object ID to filter the DataFrame by.

        Returns
        -------
        pl.DataFrame
            A filtered DataFrame containing only rows where the object_id matches.
        """
        assert hasattr(self, "plexos_data"), "plexos data not processed yet"
        return self._get_model_data(pl.col("object_id") == object_id)

    def _get_nested_object_data(self, object_id: int) -> str | float | np.ndarray:
        assert object_id
        logger.trace("Unnesting nested object", object_id)
        nested_object_id = self._resolve_object_id(object_id)
        nested_object_data = self._filter_by_object_id(nested_object_id)
        nested_object_records = nested_object_data.to_dicts()
        if len(nested_object_records) > 1:
            logger.warning("Multiple nested objects")
            key_str = "text_class_name"
            if all(
                record.get(key_str) == ClassEnum.Timeslice
                for record in nested_object_records
                if key_str in record
            ):
                property_unit = {record["property_unit"] for record in nested_object_records}
                if len(property_unit) > 1:
                    raise NotImplementedError

                unit = get_pint_unit(property_unit.pop())
                timeslice_patterns = [
                    {"pattern": record.get("text"), "value": ureg.Quantity(record["property_value"], unit)}
                    for record in nested_object_records
                ]

                return time_slice_handler(
                    records=timeslice_patterns,
                    hourly_time_index=self.hourly_time_index,
                )

        nested_object_record = nested_object_records[0]  # Get the only element of the list

        match nested_object_record["child_class_name"]:
            case ClassEnum.DataFile:
                return nested_object_record["text"]
            case ClassEnum.Variable:
                return nested_object_record["property_value"]
            case _:
                raise NotImplementedError



"""Compilation of functions used on the PLEXOS parser."""
# ruff: noqa

from datetime import datetime
import re
from enum import Enum
from typing import Any
from collections.abc import Sequence
from os import PathLike
from pathlib import Path

from numpy._typing import NDArray
import pint
import polars as pl
import numpy as np
from loguru import logger

PLEXOS_ACTION_MAP = {
    "×": np.multiply,  # noqa
    "+": np.add,
    "-": np.subtract,
    "/": np.divide,
    "=": lambda x, y: y,
}


class DATAFILE_COLUMNS(Enum):  # noqa: N801
    """Enum of possible Data file columns in Plexos."""

    NV = ("name", "value")
    Y = ("year",)
    YV = ("year", "value")
    TS_Datetime = ("datetime",)
    PV = ("pattern", "value")
    TS_NPV = ("name", "pattern", "value")
    TS_NYV = ("name", "year", "value")
    TS_NDV = ("name", "DateTime", "value")
    TS_YMDP = ("year", "month", "day", "period")
    TS_YMDPV = ("year", "month", "day", "period", "value")
    TS_NYMDV = ("name", "year", "month", "day", "value")
    TS_NYMDPV = ("name", "year", "month", "day", "period", "value")
    TS_YM = ("year", "month")
    TS_MDP = ("month", "day", "period")
    TS_NMDP = ("name", "month", "day", "period")
    TS_YMDH = (
        "year",
        "month",
        "day",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
        "19",
        "20",
        "21",
        "22",
        "23",
        "24",
    )
    TS_NYMDH = (
        "name",
        "year",
        "month",
        "day",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
        "19",
        "20",
        "21",
        "22",
        "23",
        "24",
    )
    TS_NMDH = (
        "name",
        "month",
        "day",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
        "19",
        "20",
        "21",
        "22",
        "23",
        "24",
    )
    TS_NM = (
        "name",
        "m01",
        "m02",
        "m03",
        "m04",
        "m05",
        "m06",
        "m07",
        "m08",
        "m09",
        "m10",
        "m11",
        "m12",
    )


def get_column_enum(columns: list[str]) -> DATAFILE_COLUMNS | None:
    """Identify the corresponding PropertyColumns enum based on the given columns.

    Parameters
    ----------
    columns : List[str]
        The list of columns to inspect.

    Returns
    -------
    Optional[PropertyColumns]
        The corresponding enum if a match is found; otherwise, None.
    """
    best_match = None
    max_columns_matched = 0

    columns_set = set(columns)
    for property_type in DATAFILE_COLUMNS:
        enum_columns_set = set(property_type.value)

        if enum_columns_set.issubset(columns_set):
            if len(enum_columns_set) > max_columns_matched:
                best_match = property_type
                max_columns_matched = len(enum_columns_set)

    msg = "Matched columns = {} to property_type = {}"
    logger.trace(msg, columns, best_match)
    return best_match


def filter_property_dates(system_data: pl.DataFrame, study_year: int):
    """Filter query by date_from and date_to."""
    # Convert date_from and date_to to datetime
    system_data = system_data.with_columns(
        [
            pl.col("date_from").str.strptime(pl.Datetime, "%Y-%m-%dT%H:%M:%S").cast(pl.Date),
            pl.col("date_to").str.strptime(pl.Datetime, "%Y-%m-%dT%H:%M:%S").cast(pl.Date),
        ]
    )

    # Create two new columns year_from and year_to
    system_data = system_data.with_columns(
        [
            pl.col("date_from").dt.year().alias("year_from"),
            pl.col("date_to").dt.year().alias("year_to"),
        ]
    )

    # Remove Property by study year & date_from/to
    date_filter = ((pl.col("date_from").is_null()) | (pl.col("year_from") <= study_year)) & (
        (pl.col("date_to").is_null()) | (pl.col("year_to") >= study_year)
    )
    system_data = system_data.filter(date_filter)
    return system_data


def parse_data_file(column_type: DATAFILE_COLUMNS, data_file):
    match column_type:
        case column_type.Y:
            data_file = parse_y(data_file)
        case column_type.YV:
            data_file = parse_yv(data_file)
        case column_type.TS_Datetime:
            data_file = parse_ts_datetime(data_file)
        case column_type.PV:
            data_file = parse_pv(data_file)
        case column_type.NV:
            data_file = parse_nv(data_file)
        case column_type.TS_MDP:
            data_file = parse_ts_mdp(data_file)
        case column_type.TS_NMDP:
            data_file = parse_ts_nmdp(data_file)
        case column_type.TS_NM:
            data_file = parse_ts_nm(data_file)
        case column_type.TS_YM:
            data_file = parse_ts_ym(data_file)
        case column_type.TS_YMDPV:
            data_file = parse_ts_ymdpv(data_file)
        case column_type.TS_YMDH:
            data_file = parse_ts_ymdh(data_file)
        case column_type.TS_NYMDV:
            data_file = parse_ts_nymdv(data_file)
        case column_type.TS_NYMDPV:
            data_file = parse_ts_nymdpv(data_file)
        case column_type.TS_NMDH:
            data_file = parse_ts_nmdh(data_file)
        case column_type.TS_NYMDH:
            data_file = parse_ts_nymdh(data_file)
        case _:
            msg = f"Time series format {column_type.value} not yet supported."
            raise NotImplementedError(msg)
    return data_file


def parse_y(data_file):
    data_file = data_file.melt(id_vars="year", variable_name="name")
    return data_file


def parse_yv(data_file):
    return data_file


def parse_ts_datetime(data_file):
    data_file = data_file.with_columns(datetime=pl.col("datetime").str.to_datetime("%Y-%m-%dt%H:%M"))
    data_file = data_file.with_columns(year=pl.col("datetime").dt.year())
    return data_file


def parse_pv(data_file):
    data_file = data_file.with_columns(
        month=pl.col("pattern").str.extract(r"(\d{2})$").cast(pl.Int8)
    )  # If other patterns exist, this will need to change.
    return data_file


def parse_ts_nm(data_file):
    data_file = data_file.melt(id_vars="name", variable_name="month")
    data_file = data_file.with_columns(month=pl.col("month").str.extract(r"(\d{2})$").cast(pl.Int8))
    return data_file


def parse_nv(data_file):
    return data_file


def parse_ts_nymdv(data_file):
    return data_file


def parse_ts_nymdpv(data_file):
    data_file = data_file.with_columns(hour=pl.col("period"))
    data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))
    return data_file


def parse_ts_nmdh(data_file):
    data_file = data_file.melt(id_vars=["name", "month", "day"], variable_name="hour")
    return data_file


def parse_ts_ym(data_file):
    data_file = data_file.melt(id_vars=["year", "month"], variable_name="name")
    return data_file


def parse_ts_ymdpv(data_file):
    data_file = data_file.with_columns(hour=pl.col("period"))
    data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))
    return data_file


def parse_ts_nmdp(data_file):
    data_file = data_file.with_columns(hour=pl.col("period"))
    data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))
    return data_file


def parse_ts_mdp(data_file):
    data_file = data_file.melt(id_vars=["month", "day", "period"], variable_name="name")
    # Valid PLEXOS period ranges:
    # - 1-24 periods is 1 hour data for full day
    # - 1-48 periods is 30 minute data for full day
    # - 1-289 periods is 5 minute data for full day
    if not all(data_file["period"].unique().is_in(range(0, 289))):
        msg = (
            "Periods should range 1-24 (if hourly), 1-48 (if 30min) or 1-289 (if 5min).",
            f"Found {max(data_file.unique())}",
        )
        raise NotImplementedError(msg)
    data_file = data_file.with_columns(hour=pl.col("period"))
    data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))
    return data_file


def parse_ts_ymdh(data_file):
    data_file = data_file.melt(id_vars=["year", "month", "day"], variable_name="hour")
    data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))
    data_file = data_file.sort(["year", "month", "day", "hour"])
    return data_file


def parse_ts_nymdh(data_file):
    data_file = data_file.melt(id_vars=["name", "year", "month", "day"], variable_name="hour")
    data_file = data_file.with_columns(pl.col("hour").cast(pl.Int8))
    return data_file


def parse_patterns(key: str) -> list[tuple[str, list[int]]]:
    """Parse a key for time slice patterns (e.g., 'M1-3', 'H1-6') and return a list of tuples.

    Parameters
    ----------
    key : str
        A string pattern representing time slices, such as months ('M1-12'), hours ('H1-24'),
        weekdays ('W1-7'), and days of the month ('D1-31').

    Returns
    -------
    List[tuple[str, List[int]]]
        A list of tuples where the first element is the time slice type (e.g., 'M', 'H', 'W', 'D'),
        and the second element is the list of integers representing the range of values for that time slice.

    Raises
    ------
    TypeError
        If the input is not a string.
    ValueError
        If the ranges are invalid (e.g., 'M13', 'H25').

    Examples
    --------
    >>> parse_patterns("M1-3")
    [('M', [1, 2, 3])]

    >>> parse_patterns("H1-6,H18-24")
    [('H', [1, 2, 3, 4, 5, 6]), ('H', [18, 19, 20, 21, 22, 23, 24])]

    >>> parse_patterns("W1,H1-6")
    [('W', [1]), ('H', [1, 2, 3, 4, 5, 6])]

    """
    if not isinstance(key, str):
        raise TypeError(f"Expected 'key' to be a str, got {type(key).__name__}")

    ranges = key.split(";")
    pattern_list = []

    for rng in ranges:
        time_slice_matches = re.finditer(r"([MWHD])(\d+)(?:-(\d+))?", rng)
        for match in time_slice_matches:
            time_slice_type = match.group(1)
            start_value = int(match.group(2))
            end_value = int(match.group(3)) if match.group(3) else start_value

            # Validating ranges based on time slice type
            if time_slice_type == "M" and not (1 <= start_value <= 12 and 1 <= end_value <= 12):
                raise ValueError(f"Invalid month range: {start_value}-{end_value}")
            if time_slice_type == "H" and not (1 <= start_value <= 24 and 1 <= end_value <= 24):
                raise ValueError(f"Invalid hour range: {start_value}-{end_value}")
            if time_slice_type == "W" and not (1 <= start_value <= 7 and 1 <= end_value <= 7):
                raise ValueError(f"Invalid weekday range: {start_value}-{end_value}")
            if time_slice_type == "D" and not (1 <= start_value <= 31 and 1 <= end_value <= 31):
                raise ValueError(f"Invalid day of month range: {start_value}-{end_value}")

            pattern_list.append((time_slice_type, list(range(start_value, end_value + 1))))

    return pattern_list


def time_slice_handler(
    records: list[dict[str, Any]],
    hourly_time_index: pl.DataFrame | NDArray[np.datetime64] | Sequence[datetime],
    pattern_key: str = "pattern",
) -> np.ndarray:
    """Deconstruct a dict of time slices and return a NumPy array representing a time series.

    Parameters
    ----------
    records : dist[str, Any]
        A list of dictionaries containing timeslice records.
    hourly_time_index : pl.DataFrame | NDArray[np.datetime64] | Sequence[datetime]
        Dataframe containing a 'datetime' column for hourly time index.
    pattern_key : str, optional
        Key used to extract patterns from records (default is 'pattern').

    Returns
    -------
    np.ndarray
        A NumPy array representing the time series based on the input patterns.

    Raises
    ------
    TypeError
        If records are not a list or hourly_time_index is not a polars DataFrame.
    ValueError
        If the 'datetime' column is missing from hourly_time_index.
    NotImplementedError
        If records do not contain valid month patterns starting with 'M'.

    Examples
    --------
    >>> from datetime import datetime, timedelta
    >>> records = [{"pattern": "M1-2", "value": 200}, {"pattern": "M3-12", "value": 100}]
    >>> start = datetime(year, 1, 1)
    >>> end = datetime(year + 1, 1, 1)
    >>> delta = timedelta(hours=1)
    >>> datetime_index = tuple(start + i * delta for i in range((end - start) // delta))
    >>> time_slice_handler(records, datetime_index)
    """
    if isinstance(hourly_time_index, pl.DataFrame):
        hourly_time_index = hourly_time_index.to_numpy()

    if not all(isinstance(record, dict) for record in records):
        raise TypeError("All records must be dictionaries")

    if not all(record[pattern_key].startswith("M") for record in records if pattern_key in record):
        raise NotImplementedError("All records must contain valid month patterns starting with 'M'")

    if isinstance(hourly_time_index, np.ndarray):
        hourly_time_index = hourly_time_index.astype(datetime).flatten().tolist()

    months = np.array([dt.month for dt in hourly_time_index])
    # hours = np.array([dt.hour for dt in hourly_time_index])
    month_datetime_series = np.zeros(len(hourly_time_index), dtype=float)

    for record in records:
        patterns = parse_patterns(record[pattern_key])
        for pattern in patterns:
            match pattern[0]:
                case "M":
                    month_datetime_series[np.isin(months, pattern[1])] = (
                        record["value"].magnitude
                        if isinstance(record["value"], pint.Quantity)
                        else record["value"]
                    )
                case _:
                    raise NotImplementedError

    return month_datetime_series


def find_xml(directory: PathLike):
    """
    Parameters
    ----------

    directory: Pathlike
        directory to search in for an xml file

    Raises
    ------

    FileNotFoundError:
        if no xml file is in <directory>
        OR more than one xml files are in <directory>

    NotADirectoryError:
        if <directory> is not a directory

    Returns
    -------
    ret: Path
        Path to a single xml file

    """
    directory = Path(directory)
    if not directory.is_dir():
        raise NotADirectoryError(f"Can't search {directory}, not a directory")

    xml_files = list(directory.glob("*.xml"))

    if len(xml_files) == 0:
        raise FileNotFoundError(f"No xml file in {directory}")

    if len(xml_files) > 1:
        raise FileNotFoundError(f"More than one xml file in {directory}")

    return xml_files[0]



"""Configuration handler.

This script provides the base class to save the metadata for a given ReEDS scenario.
It can either read the information directly or throught a cases file.
"""

import inspect
import os
from collections import ChainMap
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import rich
from loguru import logger
from rich.table import Table

from r2x.config_utils import get_input_defaults, get_model_config_class, get_output_defaults
from r2x.utils import get_enum_from_string, update_dict

from .config_models import BaseModelConfig, Models


@dataclass
class Scenario:
    """Scenario class.

    The purpose of the scenario class is to provide a way to track
    the inputs and defaults for a given translation. It serves as an access
    point to get the defaults, configuration and features flags for different
    steps (e.g., parser, exporter or validation) since most of the codebase use
    object composition.

    Each scenario can have its own set of flags that will run different parts
    of the parsers.

    A very simple use case for this class:

        from r2x.config_scenario import Scenario

        scenario = Scenario(name="test")
        print(scenario)

    Attributes
    ----------
    name
        Name of the translation scenario
    run_folder
        Path to input model files
    output_folder
        Path for placing the output translation files
    input_model
        Model to translate from
    output_model
        model to translate to
    feature_flags
        Dictionary with experimental features
    plugins
        List of plugins enabled
    user_dict
        Dictionary with user configuration.

    See Also
    --------
    Configuration
    """

    name: str | None = None
    run_folder: Path | str | None = None
    output_folder: Path | str = Path(".")
    input_model: str | None = None
    output_model: str | None = None
    input_config: BaseModelConfig | None = None
    output_config: BaseModelConfig | None = None
    feature_flags: dict[str, Any] = field(default_factory=dict)
    plugins: list[Any] | None = None
    user_dict: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        self._normalize_path()
        self._set_scenario_name()
        self._load_plugins()
        self._load_model_config()

        # Overload default initialization of the translation scenario if the user pass a `user_dict`. We allow
        # overriding the `input_config.defaults` and/or `output_config.defaults`. Also, we allow the user
        # to change the default file mapping by passing a `fmap` key and passing new key value pairs for
        # predetermined files that we read for each model (see `{model}_fmap.json`}. The override only happens
        # if the exact key appears on the `user_dict.
        if self.user_dict and self.input_config:
            self.input_config.defaults = update_dict(self.input_config.defaults, self.user_dict)
            self.input_config.fmap = update_dict(self.input_config.fmap, self.user_dict.get("fmap", {}))
        if self.user_dict and self.output_config:
            self.output_config.defaults = update_dict(self.output_config.defaults, self.user_dict)
        return None

    def __len__(self) -> int:
        """Return 1 scenario."""
        return 1

    def _normalize_path(self) -> None:
        if isinstance(self.output_folder, str):
            self.output_folder = Path(self.output_folder)

        # Create output folder if it does not exists.
        logger.trace(f"Creating folder {self.output_folder=}.")
        self.output_folder.mkdir(exist_ok=True)
        return

    def _set_scenario_name(self) -> None:
        if not self.name and self.run_folder:
            self.name = os.path.basename(self.run_folder)
        return None

    def _load_plugins(self) -> None:
        logger.debug("Using the following plugins: {}", self.plugins)
        return None

    def _load_model_config(self) -> None:
        if self.input_model:
            self._input_model_enum = get_enum_from_string(self.input_model, Models)
            self.input_config = get_model_config_class(self._input_model_enum)
            self.input_config.defaults = get_input_defaults(self._input_model_enum)
        if self.output_model:
            self._output_model_enum = get_enum_from_string(self.output_model, Models)
            self.output_config = get_model_config_class(self._output_model_enum)
            self.output_config.defaults = get_output_defaults(self._output_model_enum)
        return None

    def info(self) -> None:
        """Return table summary of the configuration."""
        config_table = Table(
            title="R2X Scenario Configuration",
            show_header=True,
            title_justify="left",
            title_style="bold",
        )
        config_table.add_column("Property", style="green", justify="left", min_width=20)
        config_table.add_column("Value", justify="left", min_width=20)

        for _field in self.__dict__:
            value = getattr(self, _field)
            if _field == "fmap":
                value = f"Length: {len(value)}"
            elif _field == "defaults":
                value = f"Length: {len(value)}"
            if value:
                config_table.add_row(_field, str(value))

        return rich.print(config_table)

    @classmethod
    def from_kwargs(
        cls, input_model: str, output_model: str, user_dict: dict | None = None, **kwargs
    ) -> "Scenario":
        """Create Scenario instance from key arguments."""
        cls_fields = {field for field in inspect.signature(cls).parameters}

        input_model_enum = get_enum_from_string(input_model, Models)
        input_config = get_model_config_class(input_model_enum)

        output_model_enum = get_enum_from_string(output_model, Models)
        output_config = get_model_config_class(output_model_enum)

        input_config_fields = {field for field in input_config.model_fields}
        output_config_fields = {field for field in output_config.model_fields}

        native_args, input_config_args, output_config_args, other_args = {}, {}, {}, {}
        for name, val in kwargs.items():
            if name in cls_fields:
                native_args[name] = val
            elif name in input_config_fields:
                input_config_args[name] = val
            elif name in output_config_fields:
                output_config_args[name] = val
            else:
                other_args[name] = val

        instance = cls(input_model=input_model, output_model=output_model, user_dict=user_dict, **native_args)

        # Set configurations if additional args were passed.
        if input_config_args:
            for name, value in input_config_args.items():
                setattr(instance.input_config, name, value)
        if output_config_args:
            for name, value in output_config_args.items():
                setattr(instance.output_config, name, value)

        for new_name, new_val in other_args.items():
            setattr(instance, new_name, new_val)
        return instance


@dataclass
class Configuration:
    """Configuration manager that wraps multiple Scenario instances.

    Attributes
    ----------
    scenarios
        Dictionary of scenarios to be translated.

    See Also
    --------
    Scenario
    """

    scenarios: dict[str, Scenario] = field(default_factory=dict)

    def __len__(self) -> int:
        """Get the number of scenarios in the class."""
        return len(self.scenarios)

    def __getitem__(self, scenario_name: str) -> Scenario:
        """Return a single scenario from the configuration.

        Parameters
        ----------
        name
            Name of the scenario

        Returns
        -------
            Scenario

        Raises
        ------
        KeyError
            If scenario does not exists.
        """
        for scenario in self.scenarios:
            if scenario == scenario_name:
                return self.scenarios[scenario]
        raise KeyError(f"No scenario named '{scenario_name}'")

    def __iter__(self):
        return iter(self.scenarios)

    def list_scenarios(self):
        """Return a list of scenarios in the configuration."""
        return self.scenarios.keys()

    @classmethod
    def from_cli(cls, cli_args: dict[str, Any], user_dict: dict[str, Any] | None = None, **kwargs):
        """Create a `Scenario` from the CLI arguments.

        It saves the created scenario using CLI input. It can also be overwrited if an `user_dict` is passed.

        Parameters
        ----------
        cli_args
            Arguments for constructing the scenario.
        user_dict
            Configuration for the translation.
        kwargs
            Arguments for constructing the scenario.

        See Also
        --------
            Scenario.from_kwargs
        """
        instance = cls()

        mapping = ChainMap(cli_args, user_dict or {}, kwargs)
        scenario = Scenario.from_kwargs(user_dict=user_dict, **mapping)
        assert scenario.name
        instance.scenarios[scenario.name] = scenario
        return instance

    @classmethod
    def from_scenarios(cls, cli_args: dict, user_dict: dict):
        """Create scenario from scenarios on the config file.

        This method takes the `user_dict['scenarios'] key which is a list of dicts to create the different
        scenarios for translation. The order of override is CLI -> Global Keys -> Scenario Keys

        Parameters
        ----------
        cli_args
            Arguments for constructing the scenario.
        user_dict
            Configuration for the translation.

        See Also
        --------
            Scenario.from_kwargs
        """
        instance = cls()

        global_keys = {key: value for key, value in user_dict.items() if key != "scenarios"}

        for scenario_dict in user_dict["scenarios"]:
            user_dict_scenario = {}
            if "fmap" in global_keys:
                user_dict_scenario["fmap"] = global_keys.pop("fmap")
            scenario_choices = ChainMap(cli_args, global_keys, scenario_dict)
            scenario_class = Scenario.from_kwargs(user_dict=user_dict_scenario, **scenario_choices)
            assert scenario_class.name
            instance.scenarios[scenario_class.name] = scenario_class
        return instance


def get_scenario_configuration(cli_args: dict, user_dict: dict | None = None) -> Configuration:
    """Create configuration class using CLI arguments."""
    if solve_year := cli_args.get("solve_year"):
        cli_args["solve_year"] = solve_year[0] if len(solve_year) == 1 else solve_year

    if user_dict is not None and "scenarios" in user_dict:
        config = Configuration.from_scenarios(cli_args, user_dict)
    else:
        config = Configuration.from_cli(cli_args, user_dict=user_dict)
    assert config

    # if scenario_name := cli_args.get("scenario_name", None):
    #     config = config[scenario_name]
    return config
